import csv
import errno
import itertools
import re
import shutil
import tempfile
import warnings

from pathlib import Path

try:
    # use a faster json package by default
    import ujson as json
except ImportError:  # pragma: no cover -- can't test both branches in the same run
    # fallback to the stdlib version
    import json

from katalytic.data import as_list_of_lists, sort_dict_by_keys, is_none_of
from katalytic.meta import extract_call_stack_info
# noinspection PyProtectedMember
from katalytic._pkg import _UNDEFINED, find_functions_marked_with, mark, KatalyticInterrupt


@mark('load::csv')
def load_csv(path, *, default=_UNDEFINED, encoding='utf-8'):
    """
    Load data from a CSV file and return it as a list of dictionaries.
    It will guess the data type of each column based on its values.

    Args:
        path (str):
            The path to the CSV file.
        default (Any):
            The default value to return if the specified file path does not exist.
        encoding (str, optional):
            The encoding of the CSV file. Defaults to 'utf-8'.

    Returns:
        list:
            A list of dictionaries, where each dictionary corresponds to a row
            and the keys are the column headers. The values are converted to
            appropriate data types based on their contents, with empty strings
            converted to None.
    """
    # Using the wrong extension would return default or raise an error even if the file exists.
    # Trigger the warning before that happens, to provide a hint to the user about the real problem
    _warn_if_another_function_should_be_used(path, _load_funcs)
    if not Path(path).exists() and default is not _UNDEFINED:
        return default

    with open(path, 'r', encoding=encoding) as f:
        dialect = csv.Sniffer().sniff(f.read(1024))
        f.seek(0)

        reader = csv.reader(f, dialect)
        header = next(reader)
        data = [dict(zip(header, row)) for row in reader]

    for col in header:
        t = _guess_type(col, data)

        for i, row in enumerate(data):
            # The csv reader reads missing values as empty strings.
            # Convert all empty strings to None even though they might
            # be actual empty strings
            if row[col] in ('None', ''):
                data[i][col] = None

        if t == 'bool':
            for i, row in enumerate(data):
                if row[col] == 'True':
                    data[i][col] = True
                elif row[col] == 'False':
                    data[i][col] = False
        elif t == 'float':
            for i, row in enumerate(data):
                if row[col] is not None:
                    data[i][col] = float(row[col])
        elif t == 'int':
            for i, row in enumerate(data):
                if row[col] is not None:
                    data[i][col] = int(row[col])

    return data


def _guess_type(col, data):
    """
    Guesses the data type of a column in a dataset based on its values.

    Args:
        col (str):
            The column name to guess the data type for.
        data (list of dict):
            The dataset in the form of a list of dictionaries, where each dictionary
            represents a row and with the header as keys.

    Returns:
        str:
            The guessed data type of the column. Possible values are 'bool', 'float',
            'int', or 'str'.
    """
    is_bool = True
    is_float = False
    is_number = True

    for i, row in enumerate(data):
        v = row[col]
        if not (is_number or is_bool):
            return 'str'

        if v in ('None', ''):
            continue

        if v in ('True', 'False'):
            is_number = False
            continue
        else:
            is_bool = False

        if not re.search(r'^-?\d+(\.\d+)?$', v):
            is_number = False
        elif '.' in v:
            is_float = True

    if is_bool:
        return 'bool'
    elif is_number:
        if is_float:
            return 'float'
        else:
            return 'int'
    else:
        return 'str'


@mark('save::csv')
def save_csv(data, path, *, encoding='utf-8', exists='replace', make_dirs=True):
    """
    Save data to a CSV file.

    Args:
        data (iterable):
            The data to be saved. It can be any of the following formats:
            a dict of lists, a list of lists, a list of dicts
        path (str):
            The path to the CSV file.
        encoding (str, optional):
            The encoding of the CSV file. Defaults to 'utf-8'.
        exists (str, optional):
            Specifies the behavior if the destination file already exists. Defaults to 'replace'.

            - 'error': Raise an error.
            - 'replace': Replace the existing file.
            - 'skip': Skip copying the file.
        make_dirs (bool or str, optional):
            Specifies whether to create the destination directory if it doesn't exist. Defaults to True.

            - True: Create the directory if it doesn't exist.
            - False: Raise an error if the destination directory doesn't exist.

    Returns:
        None

    Raises:
        IOError:
            If there is an error while saving the CSV file.
    """
    if exists not in ('error', 'replace', 'skip'):
        raise ValueError(f'<exists> expects "error", "replace", or "skip". Got {exists!r}')
    elif not isinstance(make_dirs, bool):
        raise TypeError(f'<make_dirs> expects False, or True. Got {make_dirs!r}')

    # optimization: trigger the warning only if the parameter preconditions are valid
    _warn_if_another_function_should_be_used(path, _save_funcs)

    # optimization: run the checks before starting to save the file, just in case the
    # path exists already when the function is called. This will ensure we don't
    # waste time saving the file only to raise an error or skip it when we do the same checks
    # right at the end of the function
    if Path(path).exists():
        if exists == 'error':
            raise FileExistsError(f'[Errno {errno.EEXIST}] File exists: {str(path)!r}')
        elif exists == 'replace':
            pass  # continue executing
        elif exists == 'skip':
            return

    dest_dir = Path(path).parent
    if make_dirs:
        make_dir(dest_dir, create_parents=True, exists_ok=True)
    elif not dest_dir.exists():
        raise FileNotFoundError(f'[Errno {errno.ENOENT}] Directory does not exist: {str(dest_dir)!r}')
    elif not dest_dir.is_dir():
        raise NotADirectoryError(f'[Errno {errno.ENOTDIR}] Not a directory: {str(dest_dir)!r}')

    data = as_list_of_lists(data)

    try:
        tmp_path = f'{path}.part'
        with open(tmp_path, 'w', newline='', encoding=encoding) as f:
            csv.writer(f, quoting=csv.QUOTE_ALL).writerows(data)

        # You could move the atomicity code higher in the function, but then
        # you wouldn't be testing the function for the worst case scenario
        if save_csv.__katalytic_test_atomicity_race_condition__:
            save_csv.__katalytic_test_atomicity_race_condition__ = False

            # I can't use save_csv([{'race': 'condition'}], path) directly
            # It would replace the tmp_path = f'{path}.part' created above
            # and then move it to the target `path`. This function wouldn't
            # be able to find the tmp_path anymore and will throw an error
            # at the end of the function: `Path(tmp_path).rename(path)`
            tmp_path_2 = path.replace('.csv', '.2.csv')
            save_csv([{'race': 'condition'}], tmp_path_2)
            Path(tmp_path_2).rename(path)

        # Checking these conditions again to make the function
        # as robust as possible against race conditions
        if Path(path).exists():
            if exists == 'error':
                raise FileExistsError(f'[Errno {errno.EEXIST}] File exists: {str(path)!r}')
            elif exists == 'replace':
                pass  # continue executing
            elif exists == 'skip':
                return

        if save_csv.__katalytic_test_atomicity_interrupt__:
            save_csv.__katalytic_test_atomicity_interrupt__ = False
            raise KatalyticInterrupt(f'Testing atomicity ...')

        # The rename is atomic on POSIX systems, but not guaranteed on Windows
        Path(tmp_path).rename(path)
    except BaseException as e:
        if not isinstance(e, KatalyticInterrupt):
            raise



@mark('load::json')
def load_json(path, *, default=_UNDEFINED, encoding='utf-8'):
    """
    Load data from a JSON file.

    Args:
        path (str):
            The path to the JSON file.
        default (Any):
            The default value to return if the specified file path does not exist.
        encoding (str, optional):
            The encoding of the JSON file. Defaults to 'utf-8'.

    Returns:
        dict or list:
            The loaded JSON data as a dictionary or list, depending on the JSON structure.

    Raises:
        FileNotFoundError:
            If the specified file path does not exist.
        IOError:
            If there is an error while reading the JSON file.
        JSONDecodeError:
            If the JSON file is not properly formatted and cannot be decoded.

    """
    # Using the wrong extension would return default or raise an error even if the file exists.
    # Trigger the warning before that happens, to provide a hint to the user about the real problem
    _warn_if_another_function_should_be_used(path, _load_funcs)
    if not Path(path).exists() and default is not _UNDEFINED:
        return default

    with open(path, encoding=encoding) as f:
        return json.load(f)


@mark('save::json')
def save_json(data, path, *, encoding='utf-8', exists='replace', indent=4, make_dirs=True, sort_keys=True):
    """
    Save data to a JSON file.

    Args:
        data (dict or list):
            The data to be saved as JSON. It should be a dictionary or a list.
        path (str):
            The path to the output JSON file.
        encoding (str, optional):
            The encoding of the JSON file. Defaults to 'utf-8'.
        exists (str, optional):
            Specifies the behavior if the destination file already exists. Defaults to 'replace'.

            - 'error': Raise an error.
            - 'replace': Replace the existing file.
            - 'skip': Skip copying the file.
        indent (int or None, optional):
            The number of spaces used for indentation in the output JSON file.
            If indent is a non-negative integer, it specifies the number of spaces
            for each indentation level. If indent is None or not provided, the output
            JSON will be compact without any indentation. Defaults to 4.
        make_dirs (bool or str, optional):
            Specifies whether to create the destination directory if it doesn't exist. Defaults to True.

            - True: Create the directory if it doesn't exist.
            - False: Raise an error if the destination directory doesn't exist.
        sort_keys (bool, optional):
            Specifies whether the keys in the output JSON should be sorted. Defaults to True.

    Returns:
        None

    Raises:
        TypeError:
            If <indent> is not an integer or a positive integer.
        ValueError:
            If <indent> is a negative integer.
        IOError:
            If there is an error while saving the JSON file.

    """
    if isinstance(indent, float) and indent.is_integer():
        indent = int(indent)
    elif not isinstance(indent, int):
        raise TypeError(f'<indent> expects a positive integer. Got {indent!r}')
    elif isinstance(indent, bool):
        raise TypeError(f'<indent> expects a positive integer. Got {indent!r}')
    elif indent < 0:
        raise ValueError(f'<indent> expects a positive integer. Got {indent!r}')
    elif exists not in ('error', 'replace', 'skip'):
        raise ValueError(f'<exists> expects "error", "replace", or "skip". Got {exists!r}')
    elif not isinstance(make_dirs, bool):
        raise TypeError(f'<make_dirs> expects False, or True. Got {make_dirs!r}')

    # optimization: trigger the warning only if the preconditions are valid
    _warn_if_another_function_should_be_used(path, _save_funcs)
    if Path(path).exists():
        if exists == 'error':
            raise FileExistsError(f'[Errno {errno.EEXIST}] File exists: {str(path)!r}')
        elif exists == 'replace':
            pass  # continue executing
        elif exists == 'skip':
            return

    dest_dir = Path(path).parent
    if make_dirs:
        make_dir(dest_dir, create_parents=True, exists_ok=True)
    elif not dest_dir.exists():
        raise FileNotFoundError(f'[Errno {errno.ENOENT}] Directory does not exist: {str(dest_dir)!r}')
    elif not dest_dir.is_dir():
        raise NotADirectoryError(f'[Errno {errno.ENOTDIR}] Not a directory: {str(dest_dir)!r}')

    try:
        tmp_path = f'{path}.part'
        with open(tmp_path, 'w', encoding=encoding) as f:
            json.dump(data, f, indent=indent, sort_keys=sort_keys)

        # You could move the atomicity code higher in the function, but then
        # you wouldn't be testing the function for the worst case scenario
        if save_json.__katalytic_test_atomicity_race_condition__:
            save_json.__katalytic_test_atomicity_race_condition__ = False

            # I can't use save_json([{'race': 'condition'}], path) directly
            # It would replace the tmp_path = f'{path}.part' created above
            # and then move it to the target `path`. This function wouldn't
            # be able to find the tmp_path anymore and will throw an error
            # at the end of the function: `Path(tmp_path).rename(path)`
            tmp_path_2 = path.replace('.json', '.2.json')
            save_json([{'race': 'condition'}], tmp_path_2)
            Path(tmp_path_2).rename(path)

        # Checking these conditions again to make the function
        # as robust as possible against race conditions
        if Path(path).exists():
            if exists == 'error':
                raise FileExistsError(f'[Errno {errno.EEXIST}] File exists: {str(path)!r}')
            elif exists == 'replace':
                pass  # continue executing
            elif exists == 'skip':
                return

        if save_json.__katalytic_test_atomicity_interrupt__:
            save_json.__katalytic_test_atomicity_interrupt__ = False
            raise KatalyticInterrupt(f'Testing atomicity ...')

        Path(tmp_path).rename(path)
    except BaseException as e:
        if not isinstance(e, KatalyticInterrupt):
            raise


@mark('load::txt')
def load_text(path, *, default=_UNDEFINED, encoding='utf-8'):
    """
    Load data from a text file.

    Args:
        path (str):
            The path to the text file.
        default (Any):
            The default value to return if the specified file path does not exist.
        encoding (str, optional):
            The encoding of the text file. Defaults to 'utf-8'.

    Returns:
        str:
            The text content read from the file.

    Raises:
        FileNotFoundError:
            If the specified file path does not exist.
        IOError:
            If there is an error while reading the text file.

    """
    # Using the wrong extension would return default or raise an error even if the file exists.
    # Trigger the warning before that happens, to provide a hint to the user about the real problem
    _warn_if_another_function_should_be_used(path, _load_funcs)
    if not Path(path).exists() and default is not _UNDEFINED:
        return default

    with open(path, encoding=encoding) as f:
        return f.read()


@mark('save::txt')
def save_text(data, path, *, encoding='utf-8', exists='replace', make_dirs=True):
    """
    Save data to a text file.

    Args:
        data (str):
            The text content to be saved.
        path (str):
            The path to the output text file.
        encoding (str, optional):
            The encoding of the text file. Defaults to 'utf-8'.
        exists (str, optional):
            Specifies the behavior if the destination file already exists. Defaults to 'replace'.

            - 'error': Raise an error.
            - 'replace': Replace the existing file.
            - 'skip': Skip copying the file.
        make_dirs (bool or str, optional):
            Specifies whether to create the destination directory if it doesn't exist. Defaults to True.

            - True: Create the directory if it doesn't exist.
            - False: Raise an error if the destination directory doesn't exist.

    Returns:
        int:
            The number of characters written to the file.

    Raises:
        IOError:
            If there is an error while saving the text file.

    """
    if exists not in ('error', 'replace', 'skip'):
        raise ValueError(f'<exists> expects "error", "replace", or "skip". Got {exists!r}')
    elif not isinstance(make_dirs, bool):
        raise TypeError(f'<make_dirs> expects False, or True. Got {make_dirs!r}')

    # optimization: trigger the warning only if the preconditions are valid
    _warn_if_another_function_should_be_used(path, _save_funcs)
    if Path(path).exists():
        if exists == 'error':
            raise FileExistsError(f'[Errno {errno.EEXIST}] File exists: {str(path)!r}')
        elif exists == 'replace':
            pass  # continue executing
        elif exists == 'skip':
            return

    dest_dir = Path(path).parent
    try:
        if make_dirs:
            make_dir(dest_dir, create_parents=True, exists_ok=True)
        elif not dest_dir.exists():
            raise FileNotFoundError(f'[Errno {errno.ENOENT}] Directory does not exist: {str(dest_dir)!r}')
        elif not dest_dir.is_dir():
            raise NotADirectoryError(f'[Errno {errno.ENOTDIR}] Not a directory: {str(dest_dir)!r}')

        tmp_path = f'{path}.part'
        with open(tmp_path, 'w', encoding=encoding) as f:
            f.write(data)

        # You could move the atomicity code higher in the function, but then
        # you wouldn't be testing the function for the worst case scenario
        if save_text.__katalytic_test_atomicity_race_condition__:
            save_text.__katalytic_test_atomicity_race_condition__ = False

            # I can't use save_text('race condition', path) directly
            # It would replace the tmp_path = f'{path}.part' created above
            # and then move it to the target `path`. This function wouldn't
            # be able to find the tmp_path anymore and will throw an error
            # at the end of the function: `Path(tmp_path).rename(path)`
            tmp_path_2 = path.replace('.txt', '.2.txt')
            save_text('race condition', tmp_path_2)
            Path(tmp_path_2).rename(path)

        # Checking these conditions again to make the function
        # as robust as possible against race conditions
        if Path(path).exists():
            if exists == 'error':
                raise FileExistsError(f'[Errno {errno.EEXIST}] File exists: {str(path)!r}')
            elif exists == 'replace':
                pass  # continue executing
            elif exists == 'skip':
                return

        if save_text.__katalytic_test_atomicity_interrupt__:
            save_text.__katalytic_test_atomicity_interrupt__ = False
            raise KatalyticInterrupt(f'Testing atomicity ...')

        Path(tmp_path).rename(path)
    except BaseException as e:
        if not isinstance(e, KatalyticInterrupt):
            raise


def clear_dir(path, *, create_missing=True):
    """
    Clear the contents of a directory.

    This function is not named "empty_dir" to avoid confusing it with "is_dir_empty".

    Args:
        path (str or Path):
            The path to the directory to be cleared.
        create_missing (bool, optional):
            Specifies whether to create the directory if it does not exist.
            If True, the directory will be created. If False, an error will be raised
            if the directory does not exist. Defaults to True.

    Returns:
        None

    Raises:
        TypeError:
            If <create_missing> is not a boolean value.
        OSError:
            If there is an error while deleting or creating the directory.

    """
    if not isinstance(create_missing, bool):
        raise TypeError(f'<create_missing> expects False or True. Got {type(create_missing)}')

    path = Path(path)
    delete_dir(path, missing_ok=True, non_empty_dir=True)
    if create_missing:
        make_dir(path)


def copy_dir(src, dest, *, dir_exists='merge', file_exists='replace', make_dirs=True):
    """
    Copy a directory from source to destination, including its contents.

    Args:
        src (str or Path):
            The path to the source directory.
        dest (str or Path):
            The path to the destination directory.
        dir_exists (str, optional):
            Specifies the behavior if the destination directory already exists. Defaults to 'merge'.

            - 'error': Raise an error.
            - 'merge': Merge the source and destination directories.
            - 'replace': Replace the destination directory with the source directory.
            - 'skip': Skip copying and return.
        file_exists (str, optional):
            Specifies the behavior if a file with the same name already exists
            in the destination directory. Defaults to 'replace'.

            - 'error': Raise an error.
            - 'replace': Replace the existing file.
            - 'skip': Skip copying the file.
        make_dirs (bool, optional):
            Specifies whether to create the destination directory if it doesn't exist. Defaults to True.

            - True: Create the directory if it doesn't exist.
            - False: Raise an error if the destination directory doesn't exist.

    Returns:
        None

    Raises:
        ValueError:
            If invalid values are provided for the 'dir_exists' or 'file_exists' parameters.
        TypeError:
            If the 'make_dirs' parameter is not a boolean value.
        FileNotFoundError:
            If the source directory doesn't exist or the destination directory cannot be created.
        NotADirectoryError:
            If the source or destination path is not a directory.
        FileExistsError:
            If a destination directory already exists when 'dir_exists' is set to 'error'.
        OSError:
            If there is an error while copying or deleting files/directories.

    """
    if dir_exists not in ('error', 'merge', 'replace', 'skip'):
        raise ValueError(f'<dir_exists> expects "error", "merge", "replace", "skip". Got {dir_exists!r}')
    elif file_exists not in ('error', 'replace', 'skip'):
        raise ValueError(f'<file_exists> expects "error", False, or True. Got {file_exists!r}')
    elif not isinstance(make_dirs, bool):
        raise TypeError(f'<make_dirs> expects False or True. Got {type(make_dirs)}')

    src = Path(src)
    dest = Path(dest)

    if not src.exists():
        raise FileNotFoundError(f'[Errno {errno.ENOENT}] <src> directory does not exist: {str(src)!r}')
    elif src.is_file():
        raise NotADirectoryError(f'[Errno {errno.ENOTDIR}] Expected a directory, but <src> is a file: {str(src)!r}')
    elif dest.is_file():
        raise NotADirectoryError(f'[Errno {errno.ENOTDIR}] Expected a directory, but <dest> is a file: {str(dest)!r}')
    elif dest.is_dir():
        if src.samefile(dest):
            raise ValueError(f'<src> and <dest> are equal: {str(src)!r}')
        elif dir_exists == 'replace':
            delete_dir(dest, non_empty_dir=True)
        elif dir_exists == 'skip':
            return

        for src_item in src.iterdir():
            dest_item = str(src_item).replace(str(src), str(dest))
            if src_item.is_dir():
                if Path(dest_item).is_dir() and dir_exists == 'error':
                    raise FileExistsError(f'[Errno {errno.EEXIST}] Directory already exists: {str(dest_item)!r}')

                copy_dir(src_item, dest_item, make_dirs=make_dirs, file_exists=file_exists, dir_exists=dir_exists)
            else:
                copy_file(src_item, dest_item, exists=file_exists)
    elif make_dirs is True:
        if src.name != dest.name:
            dest = dest/src.name

        make_dir(dest)
        copy_dir(src, dest, make_dirs=make_dirs, file_exists=file_exists, dir_exists=dir_exists)
    elif make_dirs is False:
        raise FileNotFoundError(f'[Errno {errno.ENOENT}] No such directory: {str(dest)!r}')


def copy_file(src, dest, *, exists='replace', make_dirs=True):
    """
    Copy a file from source to destination.

    Args:
        src (str or Path):
            The path to the source file.
        dest (str or Path):
            The path to the destination file.
        exists (str, optional):
            Specifies the behavior if the destination file already exists. Defaults to 'replace'.

            - 'error': Raise an error.
            - 'replace': Replace the existing file.
            - 'skip': Skip copying the file.
        make_dirs (bool or str, optional):
            Specifies whether to create the destination directory if it doesn't exist. Defaults to True.

            - True: Create the directory if it doesn't exist.
            - False: Raise an error if the destination directory doesn't exist.

    Returns:
        None

    Raises:
        ValueError:
            If invalid values are provided for the 'exists' or 'make_dirs' parameters.
        FileNotFoundError:
            If the source file doesn't exist or the destination directory cannot be created.
        ValueError:
            If the source and destination paths are the same.
        IsADirectoryError:
            If the source path is a directory.
        FileExistsError:
            If a destination file already exists when 'exists' is set to 'error'.
        OSError:
            If there is an error while copying or deleting the file.

    """
    if exists not in ('error', 'replace', 'skip'):
        raise ValueError(f'<exists> expects "error", "replace", or "skip". Got {exists!r}')
    elif not isinstance(make_dirs, bool):
        raise TypeError(f'<make_dirs> expects False or True. Got {type(make_dirs)}')

    ends_with_slash = str(dest).endswith('/')
    src = Path(src)
    dest = Path(dest)

    if src.exists() and dest.exists() and src.samefile(dest):
        raise ValueError(f'<src> and <dest> are equal: {str(src)!r}')
    elif not src.exists():
        raise FileNotFoundError(f'[Errno {errno.ENOENT}] <src> file does not exist: {str(src)!r}')
    elif src.is_dir():
        raise IsADirectoryError(f'[Errno {errno.EISDIR}] Expected a file, but <src> is a directory: {str(src)!r}')

    try:
        if dest.exists():
            if exists == 'error':
                raise FileExistsError(f'[Errno {errno.EEXIST}] File exists: {str(dest)!r}')
            elif exists == 'replace':
                pass
            elif exists == 'skip':
                return
        elif make_dirs:
            if ends_with_slash:
                make_dir(dest)
            else:
                make_dir(dest.parent)
        else:
            raise FileNotFoundError(f'[Errno {errno.ENOENT}] No such file: {str(dest)!r}')

        # You could move the atomicity code higher in the function, but then
        # you wouldn't be testing the function for the worst case scenario
        if copy_file.__katalytic_test_atomicity_race_condition__:
            copy_file.__katalytic_test_atomicity_race_condition__ = False

            # I can't use copy_file([{'race': 'condition'}], path) directly
            # It would replace the tmp_path = f'{path}.part' created above
            # and then move it to the target `path`. This function wouldn't
            # be able to find the tmp_path anymore and will throw an error
            # at the end of the function: `Path(tmp_path).rename(path)`
            Path(dest).write_text('race condition')

        # Checking these conditions again to make the function
        # as robust as possible against race conditions
        if Path(dest).exists():
            if exists == 'error':
                raise FileExistsError(f'[Errno {errno.EEXIST}] File exists: {str(dest)!r}')
            elif exists == 'replace':
                pass  # continue executing
            elif exists == 'skip':
                return

        if copy_file.__katalytic_test_atomicity_interrupt__:
            copy_file.__katalytic_test_atomicity_interrupt__ = False
            raise KatalyticInterrupt()

        shutil.copy2(src, dest)
    except BaseException as e:
        if not isinstance(e, KatalyticInterrupt):
            raise


def delete_dir(path, *, missing_ok=True, non_empty_dir=True):
    """
    Delete a directory.

    Args:
        path (str or Path):
            The path to the directory to be deleted.
        missing_ok (bool, optional):
            Specifies whether it is okay if the directory doesn't exist. Defaults to True.

            - True: No error is raised if the directory doesn't exist.
            - False: Raise an error if the directory doesn't exist.
        non_empty_dir (str or bool, optional):
            Specifies the behavior if the directory is non-empty.  Defaults to True.

            - 'error': Raise an error if the directory is non-empty.
            - False: Delete the directory even if it is non-empty.
            - True: Delete the directory and its contents, recursively.

    Returns:
        None

    Raises:
        ValueError:
            If the provided path is empty or refers to the current directory.
            If invalid values are provided for the 'missing_ok' or 'non_empty_dir' parameters.
        TypeError:
            If the 'missing_ok' parameter is not a boolean value.
        FileNotFoundError:
            If the directory doesn't exist and 'missing_ok' is set to False.
        NotADirectoryError:
            If the provided path refers to a file instead of a directory.
        OSError:
            If there is an error while deleting the directory.

    """
    if path is None or path in ('', '.'):
        raise ValueError(f'path={path!r} would delete the current dir')
    elif not isinstance(missing_ok, bool):
        raise TypeError(f'<missing_ok> expects False or True. Got {type(missing_ok)}')
    elif is_none_of(non_empty_dir, ('error', False, True)):
        raise ValueError(f'<non_empty_dir> expects "error", False, or True. Got {non_empty_dir!r}')

    path = Path(path)
    if path.is_dir():
        if non_empty_dir == 'error':
            path.rmdir()
        elif non_empty_dir is True:
            shutil.rmtree(path)
        elif non_empty_dir is False and not is_dir_empty(path):
            pass
    elif not path.exists():
        if missing_ok:
            pass
        else:
            raise FileNotFoundError(f'[Errno {errno.ENOENT}] No such directory: {str(path)!r}')
    elif path.is_file():
        raise NotADirectoryError(f'[Errno {errno.ENOTDIR}] Expected a directory, but <path> is a file: {str(path)!r}')


def delete_file(path, *, missing_ok=True):
    """
    Delete a file.

    Args:
        path (str or Path):
            The path to the file to be deleted.
        missing_ok (bool, optional):
            Specifies whether it is okay if the file doesn't exist. Defaults to True.

            - True: No error is raised if the file doesn't exist.
            - False: Raise an error if the file doesn't exist.

    Returns:
        None

    Raises:
        TypeError:
            If the 'missing_ok' parameter is not a boolean value.
        FileNotFoundError:
            If the file doesn't exist and 'missing_ok' is set to False.
        IsADirectoryError:
            If the provided path refers to a directory instead of a file.
        OSError:
            If there is an error while deleting the file.

    """
    if not isinstance(missing_ok, bool):
        raise TypeError(f'<missing_ok> expects False or True. Got {type(missing_ok)}')

    path = Path(path)
    if path.is_file():
        path.unlink()
    elif not path.exists():
        if missing_ok:
            pass
        else:
            raise FileNotFoundError(f'[Errno {errno.ENOENT}] No such file: {str(path)!r}')
    elif path.is_dir():
        raise IsADirectoryError(f'[Errno {errno.EISDIR}] Expected a file, but <path> is a directory: {str(path)!r}')


def get_all(path='.', *, glob=True, iter_=False, prefix=True, recursive=False):
    """
    Retrieve all files and directories at the specified path. If `iter_` is False, they will be sorted

    Args:
        path (str or Path, optional):
            The path to retrieve files and directories from. Defaults to the current directory ('.').
        glob (bool, optional):
            Specifies whether to use glob pattern matching to retrieve files and directories. Defaults to True.
        iter_ (bool, optional):
            Specifies whether to return an iterator instead of a list of results. Defaults to False.
        prefix (bool, optional):
            Specifies whether to include the path prefix in the returned results. Defaults to True.
        recursive (bool, optional):
            Specifies whether to retrieve files and directories recursively. Defaults to False.

    Returns:
        list or iterator:
            A list of files and directories at the specified path, or an iterator if `iter_` is set to True.

    """
    return _get_all(path, glob=glob, iter_=iter_, prefix=prefix, recursive=recursive)


def get_dirs(path='.', *, glob=True, iter_=False, prefix=True, recursive=False):
    """
    Retrieve directories at the specified path. If `iter_` is False, they will be sorted

    Args:
        path (str or Path, optional):
            The path to retrieve directories from. Defaults to the current directory ('.').
        glob (bool, optional):
            Specifies whether to use glob pattern matching to retrieve directories. Defaults to True.
        iter_ (bool, optional):
            Specifies whether to return an iterator instead of a list of results. Defaults to False.
        prefix (bool, optional):
            Specifies whether to include the path prefix in the returned results. Defaults to True.
        recursive (bool, optional):
            Specifies whether to retrieve directories recursively. Defaults to False.

    Returns:
        list or iterator:
            A list of directories at the specified path, or an iterator if `iter_` is set to True.

    """
    return _get_all(path, glob=glob, iter_=iter_, prefix=prefix, only_dirs=True, recursive=recursive)


def get_files(path='.', *, glob=True, iter_=False, prefix=True, recursive=False):
    """
    Retrieve files at the specified path. If `iter_` is False, they will be sorted

    Args:
        path (str or Path, optional):
            The path to retrieve files from. Defaults to the current directory ('.').
        glob (bool, optional):
            Specifies whether to use glob pattern matching to retrieve files. Defaults to True.
        iter_ (bool, optional):
            Specifies whether to return an iterator instead of a list of results. Defaults to False.
        prefix (bool, optional):
            Specifies whether to include the path prefix in the returned results. Defaults to True.
        recursive (bool, optional):
            Specifies whether to retrieve files recursively. Defaults to False.

    Returns:
        list or iterator:
            A list of files at the specified path, or an iterator if `iter_` is set to True.

    """
    return _get_all(path, glob=glob, iter_=iter_, prefix=prefix, only_files=True, recursive=recursive)


def _get_all(path='.', *, glob=True, iter_=False, only_dirs=False, only_files=False, prefix=True, recursive=False):
    """
    Retrieve all files and directories at the specified path. If `iter_` is False, they will be sorted

    Args:
        path (str or Path, optional):
            The path to retrieve files and directories from. Defaults to the current directory ('.').
        glob (bool, optional):
            Specifies whether to use glob pattern matching to retrieve files and directories. Defaults to True.
        iter_ (bool, optional):
            Specifies whether to return an iterator instead of a list of results. Defaults to False.
        only_dirs (bool, optional):
            Specifies whether to retrieve only directories. Defaults to False.
        only_files (bool, optional):
            Specifies whether to retrieve only files. Defaults to False.
        prefix (bool, optional):
            Specifies whether to include the path prefix in the returned results. Defaults to True.
        recursive (bool, optional):
            Specifies whether to retrieve files and directories recursively. Defaults to False.

    Returns:
        list or iterator:
            A list of files and directories at the specified path, or an iterator if `iter_` is set to True.

    Raises:
        TypeError:
            If invalid values are provided for the boolean parameters.
        ValueError:
            If both 'only_dirs' and 'only_files' are set to True.

    """
    if not isinstance(glob, bool):
        raise TypeError(f'<glob> expects False or True. Got {type(glob)}')
    elif not isinstance(iter_, bool):
        raise TypeError(f'<iter_> expects False or True. Got {type(iter_)}')
    elif not isinstance(only_dirs, bool):
        raise TypeError(f'<only_dirs> expects False or True. Got {type(only_dirs)}')
    elif not isinstance(only_files, bool):
        raise TypeError(f'<only_files> expects False or True. Got {type(only_files)}')
    elif not isinstance(prefix, bool):
        raise TypeError(f'<prefix> expects False or True. Got {type(prefix)}')
    elif not isinstance(recursive, bool):
        raise TypeError(f'<recursive> expects False or True. Got {type(recursive)}')
    elif only_dirs and only_files:
        raise ValueError('<only_dirs> and <only_files> can\'t be True at the same time')

    original_type = type(path)
    if glob and (path is not None and '*' in str(path)):
        path, _, pattern = str(path).partition('*')
        if recursive:
            pattern = re.sub(r'^\*/', r'**/', f'*{pattern}')
            result = Path(path).rglob(pattern)
        else:
            result = Path(path).glob(f'*{pattern}')
    else:
        result = Path(path).iterdir()
        if recursive:
            result = itertools.chain.from_iterable(
                [p] if p.is_file()
                else itertools.chain([p], _get_all(
                    p, glob=glob, iter_=True, only_dirs=only_dirs,
                    only_files=only_files, prefix=True, recursive=recursive))
                for p in result
            )

    if only_dirs:
        result = (p for p in result if Path(p).is_dir())
    elif only_files:
        result = (p for p in result if Path(p).is_file())

    if prefix:
        result = (str(p) for p in result)
    else:
        result = (str(p).replace(f'{path}/', '', 1) for p in result)

    result = (original_type(p) for p in result)

    if iter_:
        return result
    else:
        return sorted(result)


def get_unique_path(pattern='{}'):
    """
    Generate a unique path based on the provided pattern.

    Args:
        pattern (str or Path, optional):
            The pattern for generating the unique path. Defaults to '{}'.

    Returns:
        str:
            A unique path generated based on the pattern.

    Raises:
        TypeError:
            If the 'pattern' parameter is not a string or a pathlib.Path object.
        ValueError:
            If the pattern is invalid or does not contain exactly one placeholder.
    """
    if not isinstance(pattern, (str, Path)):
        raise TypeError(f'<pattern> expects a str or pathlib.Path. Got {type(pattern)}')

    original_type = type(pattern)
    pattern = str(pattern)
    placeholders = re.findall(r'{(:((\d*)?d)?)?}', pattern)
    if len(placeholders) != 1:
        raise ValueError(
            f'Invalid pattern: {pattern!r}. You must provide exactly one placeholder, '
            + 'optionally with an integer format. Try using "{}" or "{:03d}"')

    if pattern.startswith('./'):
        pattern = pattern.partition('./')[2]

    if not pattern.startswith('/'):
        d = tempfile.mkdtemp()
        pattern = f'{d}/{pattern}'

    n = 0
    while True:
        n += 1
        path = pattern.format(n)
        if not Path(path).exists():
            return original_type(path)


def is_dir_empty(path, *, missing='error'):
    """
    Check if a directory is empty.

    Args:
        path (str or Path):
            The path to the directory.
        missing (bool or str, optional):
            Specifies the behavior if the directory doesn't exist. Defaults to 'error'.

            - True: Return True if the directory doesn't exist.
            - False: Return False if the directory doesn't exist.
            - 'error': Raise an error if the directory doesn't exist.

    Returns:
        bool:
            True if the directory is empty, False otherwise.

    Raises:
        ValueError:
            If invalid values are provided for the 'missing' parameter.
        NotADirectoryError:
            If the provided path refers to a file instead of a directory.
        FileNotFoundError:
            If the directory doesn't exist and 'missing' is set to 'error'.

    """
    if is_none_of(missing, (False, True, 'error')):
        raise ValueError(f'<missing> expects "error", False, or True. Got {missing!r}')

    path = Path(path)
    if path.is_dir():
        return list(path.iterdir()) == []
    elif path.is_file():
        raise NotADirectoryError(f'[Errno {errno.ENOTDIR}] Expected a directory, but "path" is a file: {str(path)!r}')
    else:
        if missing is False:
            return False
        elif missing is True:
            return True
        elif missing == 'error':
            raise FileNotFoundError(f'[Errno {errno.ENOENT}] No such file or directory: {str(path)!r}')


def is_file_empty(path, *, missing='error'):
    """
    Check if a file is empty.

    Args:
        path (str or Path):
            The path to the file.
        missing (bool or str, optional):
            Specifies the behavior if the file doesn't exist. Defaults to 'error'.

            - True: Return True if the file doesn't exist.
            - False: Return False if the file doesn't exist.
            - 'error': Raise an error if the file doesn't exist.

    Returns:
        bool:
            True if the file is empty, False otherwise.

    Raises:
        ValueError:
            If invalid values are provided for the 'missing' parameter.
        IsADirectoryError:
            If the provided path refers to a directory instead of a file.
        FileNotFoundError:
            If the file doesn't exist and 'missing' is set to 'error'.

    """
    if is_none_of(missing, (False, True, 'error')):
        raise ValueError(f'<missing> expects "error", False, or True. Got {missing!r}')

    path = Path(path)
    if path.is_file():
        return path.stat().st_size == 0
    elif path.is_dir():
        raise IsADirectoryError(f'[Errno {errno.EISDIR}] Expected a file, but <path> is a directory: {str(path)!r}')
    else:
        if missing is False:
            return False
        elif missing is True:
            return True
        elif missing == 'error':
            raise FileNotFoundError(f'[Errno {errno.ENOENT}] No such file or directory: {str(path)!r}')


def make_dir(path, *, create_parents=True, exists_ok=True):
    """
    Create a directory at the specified path.

    Args:
        path (str or Path):
            The path to the directory.
        create_parents (bool, optional):
            Specifies whether to create parent directories if they don't exist. If set to False
            and the parent directories don't exist, an error will be raised. Defaults to True.
        exists_ok (bool, optional):
            Specifies whether to raise an error if the directory already exists. Defaults to True.

    Returns:
        None

    Raises:
        TypeError:
            If the 'create_parents' or 'exists_ok' parameters are not boolean values.
        NotADirectoryError:
            If the provided path refers to a file instead of a directory.
        FileExistsError:
            If the directory already exists and 'exists_ok' is set to False.
        OSError:
            If there is an error while creating the directory.

    """
    if not isinstance(create_parents, bool):
        raise TypeError(f'<create_parents> expects False or True. Got {type(create_parents)}')
    elif not isinstance(exists_ok, bool):
        raise TypeError(f'<exists_ok> expects False or True. Got {type(exists_ok)}')

    path = Path(path)
    if path.is_file():
        raise NotADirectoryError(f'[Errno {errno.ENOTDIR}] Expected a directory, but "path" is a file: {str(path)!r}')

    path.mkdir(parents=create_parents, exist_ok=exists_ok)


def move_dir(src, dest, *, dir_exists='merge', file_exists='replace', make_dirs=True):
    """
    Move a directory from source to destination.

    Args:
        src (str or Path):
            The path to the source directory.
        dest (str or Path):
            The path to the destination directory.
        dir_exists (str or bool, optional):
            Specifies the behavior if the destination directory already exists. Defaults to 'merge'.

            - 'error': Raise an error.
            - 'merge': Merge the source and destination directories.
            - 'replace': Replace the destination directory with the source directory.
            - 'skip': Skip copying and return.
        file_exists (str, optional):
            Specifies the behavior if a file with the same name already exists
            in the destination directory. Defaults to 'replace'.

            - 'error': Raise an error.
            - 'replace': Replace the existing file.
            - 'skip': Skip copying the file.
        make_dirs (bool, optional):
            Specifies whether to create the destination directory if it doesn't exist. Defaults to True.

            - True: Create the directory if it doesn't exist.
            - False: Raise an error if the destination directory doesn't exist.

    Returns:
        None

    """
    if dir_exists == 'skip' and Path(dest).exists():
        return

    copy_dir(src, dest, dir_exists=dir_exists, file_exists=file_exists, make_dirs=make_dirs)
    delete_dir(src, non_empty_dir=True)


def move_file(src, dest, *, exists='replace', make_dirs=True):
    """
    Move a file from source to destination.

    Args:
        src (str or Path):
            The path to the source file.
        dest (str or Path):
            The path to the destination file.
        exists (str, optional):
            Specifies the behavior if the destination file already exists. Defaults to 'replace'.

            - 'error': Raise an error.
            - 'replace': Replace the existing file.
            - 'skip': Skip copying the file.
        make_dirs (bool or str, optional):
            Specifies whether to create the destination directory if it doesn't exist. Defaults to True.

            - True: Create the directory if it doesn't exist.
            - False: Raise an error if the destination directory doesn't exist.

    Returns:
        None

    """
    if exists == 'skip' and Path(dest).exists():
        return

    copy_file(src, dest, exists=exists, make_dirs=make_dirs)
    delete_file(src)


def _find_grouped_functions(group):
    loaders = {}
    for func_name, f, groups in find_functions_marked_with(group):
        for g in groups:
            ext = g.rpartition('::')[2]
            loaders[ext] = f

    # sort the dict with the most specific extension at the beginning
    # this makes it easier to pick ".tar.gz" over ".gz"
    return sort_dict_by_keys(loaders, condition=len, reverse=True)


_load_funcs = _find_grouped_functions('load::*')
_save_funcs = _find_grouped_functions('save::*')


def load(path, default=_UNDEFINED, **kwargs):
    """
    Load data from a file based on the file extension.

    Args:
        path (str):
            The path to the input file.
        default (Any):
            The default value to return if the specified file path does not exist.
        **kwargs:
            Additional keyword arguments to be passed to the specific load function.

    Returns:
        any:
            The loaded data from the file.

    Raises:
        RuntimeError:
            If no load function is found for the given file extension.

    """
    for ext, f in _load_funcs.items():
        if path.lower().endswith(f'.{ext}'):
            return f(path, default=default, **kwargs)

    raise RuntimeError(f'No load function found for {path!r}')


def save(data, path, *, exists='replace', make_dirs=True, **kwargs):
    """
    Save data to a file based on the file extension.

    Args:
        data (any):
            The data to be saved.
        path (str):
            The path to the output file.
        exists (str, optional):
            Specifies the behavior if the destination file already exists. Defaults to 'replace'.

            - 'error': Raise an error.
            - 'replace': Replace the existing file.
            - 'skip': Skip copying the file.
        make_dirs (bool or str, optional):
            Specifies whether to create the destination directory if it doesn't exist. Defaults to True.

            - True: Create the directory if it doesn't exist.
            - False: Raise an error if the destination directory doesn't exist.
        **kwargs:
            Additional keyword arguments to be passed to the specific save function.

    Returns:
        None

    Raises:
        RuntimeError:
            If no save function is found for the given file extension.

    """
    for ext, f in _save_funcs.items():
        if path.lower().endswith(f'.{ext}'):
            f(data, path, exists=exists, make_dirs=make_dirs, **kwargs)
            return

    raise RuntimeError(f'No save function found for {path!r}')


def _warn_if_another_function_should_be_used(path, group_dict):
    path = path.lower()
    for ext, f in group_dict.items():
        if path.endswith(f'.{ext}'):
            file, caller, line = extract_call_stack_info(depth=1)
            if group_dict[ext] != caller:
                warnings.warn(
                    f'Use "{group_dict[ext].__name__}" for ".{ext}" files instead of "{caller.__name__}".'
                    f'\n(called from {file}:{line})'
                )
            break


copy_dir.__katalytic_test_atomicity_interrupt__ = False
copy_file.__katalytic_test_atomicity_interrupt__ = False
save_csv.__katalytic_test_atomicity_interrupt__ = False
save_json.__katalytic_test_atomicity_interrupt__ = False
save_text.__katalytic_test_atomicity_interrupt__ = False

copy_dir.__katalytic_test_atomicity_race_condition__ = False
copy_file.__katalytic_test_atomicity_race_condition__ = False
save_csv.__katalytic_test_atomicity_race_condition__ = False
save_json.__katalytic_test_atomicity_race_condition__ = False
save_text.__katalytic_test_atomicity_race_condition__ = False



from pathlib import Path

import pytest as pytest
from katalytic.data import all_types_besides
from katalytic.meta import extract_call_stack_info, reference_caller_function


def _f0(depth): return reference_caller_function(depth=depth)
def _f1(depth): return _f0(depth)
def _f2(depth): return _f1(depth)


def _g0(depth): return extract_call_stack_info(depth=depth)
def _g1(depth): return _g0(depth)
def _g2(depth): return _g1(depth)


class _Nested_0:
    class Nested_1:
        def f0(depth): return extract_call_stack_info(depth=depth)
        def f1(depth): return _Nested_0.Nested_1.f0(depth)
        def f2(depth): return _Nested_0.Nested_1.f1(depth)


class Test_extract_call_stack_info:
    @pytest.mark.parametrize('caller, depth, expected', [
        (_Nested_0.Nested_1.f1, 0, _Nested_0.Nested_1.f0),
        (_Nested_0.Nested_1.f1, 1, _Nested_0.Nested_1.f1),
        (_Nested_0.Nested_1.f2, 0, _Nested_0.Nested_1.f0),
        (_Nested_0.Nested_1.f2, 1, _Nested_0.Nested_1.f1),
        (_Nested_0.Nested_1.f2, 2, _Nested_0.Nested_1.f2),
    ])
    def test_depth_ok_nested(self, caller, depth, expected):
        path, func, line = caller(depth=depth)

        assert path == str(Path(__file__).resolve())
        assert func == expected
        assert isinstance(line, int) and line > 0

    @pytest.mark.parametrize('caller, depth, expected', [
        (_g1, 0, _g0),
        (_g1, 1, _g1),
        (_g2, 0, _g0),
        (_g2, 1, _g1),
        (_g2, 2, _g2),
    ])
    def test_depth_ok(self, caller, depth, expected):
        path, func, line = caller(depth=depth)

        assert path == str(Path(__file__).resolve())
        assert func == expected
        assert isinstance(line, int) and line > 0

    def test_precondition_depth_should_be_positive(self):
        with pytest.raises(ValueError):
            _g1(-1)

    def test_depth_should_not_be_larger_than_the_stack(self):
        with pytest.raises(ValueError):
            # Use a large number because of the pytest internal calls
            _g1(100)

    @pytest.mark.parametrize('depth', all_types_besides('int'))
    def test_precondition_depth_should_be_an_integer(self, depth):
        with pytest.raises(TypeError):
            _g1(depth)


class Test_reference_caller_function:
    @pytest.mark.parametrize('caller, depth, expected', [
        (_f1, 0, _f0),
        (_f1, 1, _f1),
        (_f2, 0, _f0),
        (_f2, 1, _f1),
        (_f2, 2, _f2),
    ])
    def test_depth_ok(self, caller, depth, expected):
        assert caller(depth=depth) == expected

    def test_precondition_depth_should_be_positive(self):
        with pytest.raises(ValueError):
            _f1(-1)

    def test_depth_should_not_be_larger_than_the_stack(self):
        with pytest.raises(ValueError):
            # Use a large number because of the pytest internal calls
            _f1(100)

    @pytest.mark.parametrize('depth', all_types_besides('int'))
    def test_precondition_depth_should_be_an_integer(self, depth):
        with pytest.raises(TypeError):
            _f1(depth)



import math

from katalytic.data.checks import is_number, is_sequence, is_iterable, is_iterator

_UNDEFINED = object()


def L1(a, b):
    """
    Compute the L1 (Manhattan) distance between two values or sequences.

    Parameters:
        a: The first value or sequence.
        b: The second value or sequence.

    Returns:
        The L1 distance between the two values or sequences.

    Raises:
        ValueError: If the types of a or b are not supported or unexpected.
            - If a or b is neither a number or a sequence.
            - If a and b are sequences and
                - the sequences are empty.
                - the sequences have different lengths.
                - the sequences are nested (contains sub-sequences).

    """
    if isinstance(a, bool):
        raise ValueError(f'Got <a> = {a!r}')
    elif isinstance(b, bool):
        raise ValueError(f'Got <b> = {b!r}')
    elif isinstance(a, (int, float)) and isinstance(b, (int, float)):
        return abs(a - b)
    elif is_sequence(a) and is_sequence(b):
        if len(a) == 0 and len(b) == 0:
            raise ValueError(f'Both sequences are empty')
        elif len(a) != len(b):
            raise ValueError(f'The sequences have different lengths: {len(a)} and {len(b)}')
        elif is_sequence(a[0]) or is_sequence(b[0]):
            raise ValueError(f'Nested sequences are not supported')
        else:
            return sum(L1(ai, bi) for ai, bi in zip(a, b))
    else:
        raise ValueError(f'Unknown format: ({type(a).__name__}) {a!r} and ({type(b).__name__}) {b!r}')


def L2(a, b):
    """
    Compute the L2 (Euclidean) distance between two values or sequences.

    Parameters:
        a: The first value or sequence.
        b: The second value or sequence.

    Returns:
        The L2 distance between the two values or sequences.

    Raises:
        ValueError: If the types of a or b are not supported or unexpected.
            - If a or b is neither a number or a sequence.
            - If a and b are sequences and
                - the sequences are empty.
                - the sequences have different lengths.
                - the sequences are nested (contains sub-sequences).

    """
    if isinstance(a, bool):
        raise ValueError(f'Got <a> = {a!r}')
    elif isinstance(b, bool):
        raise ValueError(f'Got <b> = {b!r}')
    elif isinstance(a, (int, float)) and isinstance(b, (int, float)):
        return math.sqrt((a - b)**2)
    elif is_sequence(a) and is_sequence(b):
        if len(a) == 0 and len(b) == 0:
            raise ValueError(f'Both sequences are empty')
        elif len(a) != len(b):
            raise ValueError(f'The sequences have different lengths: {len(a)} and {len(b)}')
        elif is_sequence(a[0]) or is_sequence(b[0]):
            raise ValueError(f'Nested sequences are not supported')
        elif any(isinstance(ai, bool) or isinstance(bi, bool) for ai, bi in zip(a, b)):
            raise ValueError(f'Got a boolean value in one of the sequences')
        elif all(isinstance(ai, (int, float)) and isinstance(bi, (int, float)) for ai, bi in zip(a, b)):
            return math.sqrt(sum((ai - bi)**2 for ai, bi in zip(a, b)))
        else:
            raise ValueError(f'Unknown format: ({type(a).__name__}) {a!r} and ({type(b).__name__}) {b!r}')
    else:
        raise ValueError(f'Unknown format: ({type(a).__name__}) {a!r} and ({type(b).__name__}) {b!r}')


def min_max(iterable, *, default=_UNDEFINED, key=None):
    """
    Find the minimum and maximum values from the given iterable.

    Parameters:
        iterable: An iterable object from which to find the minimum and maximum values.
        default (optional): The default value to return if the iterable is empty.
            If not provided, a ValueError is raised when the iterable is empty.
        key (optional): A callable function that is used to extract a comparison key from each element in the iterable.
            If not provided, the elements themselves are used for comparison.

    Returns:
        A tuple containing the minimum and maximum values from the iterable.

    Raises:
        TypeError: If the iterable is not iterable or if the key is not None or callable.
        ValueError: If the iterable is empty and no default value is provided.

    """
    if not is_iterable(iterable):
        raise TypeError(f'<iterable> expected an iterable, got {type(iterable).__name__}')
    elif not (key is None or callable(key)):
        raise TypeError(f'<key> expected a callable, got {type(key).__name__}')

    if key is None:
        key = lambda x: x

    if is_iterator(iterable):
        iterable = list(iterable)

    if len(iterable) == 0:
        if default is _UNDEFINED:
            raise ValueError(f'Cannot get the min/max of an empty iterable')
        else:
            return default

    min_ = min(iterable, key=key)
    max_ = max(iterable, key=key)
    return (min_, max_)


def clip(x, min_=float('-inf'), max_=float('+inf')):
    """
    Clip the value x within the specified minimum and maximum bounds.

    Parameters:
        x: The value to be clipped.
        min_ (optional): The minimum bound. If not provided, negative infinity is used.
        max_ (optional): The maximum bound. If not provided, positive infinity is used.

    Returns:
        The clipped value of x, which is guaranteed to be within the specified bounds.

    Raises:
        TypeError: If x, min_, or max_ is not a number.
    """
    if not is_number(x):
        raise TypeError(f'<x> expected a number, got {type(x).__name__}')
    elif not is_number(min_):
        raise TypeError(f'<min_> expected a number, got {type(min_).__name__}')
    elif not is_number(max_):
        raise TypeError(f'<max_> expected a number, got {type(max_).__name__}')

    return min(max(x, min_), max_)



import copy
import itertools
import shutil
import tempfile
import warnings
from pathlib import Path

import pytest

from katalytic.data import all_types_besides, as_dict_of_lists, as_list_of_lists, is_iterable, is_iterator
from katalytic.data.checks import dicts_share_key_order
# noinspection PyProtectedMember
from katalytic.files import (
    _load_funcs, _save_funcs, _get_all, clear_dir, copy_dir, copy_file, delete_dir, delete_file, get_dirs, get_files,
    get_unique_path, is_dir_empty, get_all, is_file_empty, load, load_csv, load_json, load_text, make_dir, move_dir,
    move_file, save, save_csv, save_json, save_text
)


def _compute_all_mismatched_combos(group_dict):
    good_matches = group_dict.items()
    all_matches = itertools.product(group_dict.keys(), group_dict.values())
    bad_matches = (m for m in all_matches if m not in good_matches)
    return ((*m, group_dict[m[0]].__name__) for m in bad_matches)


def _create_seq_of_dicts():
    return [
        {'a': 1, 'b': 'x', 'c': True, 'd': -1.5, 'e': True},
        {'a': 2, 'b': None, 'c': False, 'd': 0.0, 'e': 'z'},
        {'a': None, 'b': 'y', 'c': '', 'd': None, 'e': 'w'},
    ]


def _create_seq_of_seq():
    return as_list_of_lists(_create_seq_of_dicts())


def _create_dict_of_seq():
    return as_dict_of_lists(_create_seq_of_dicts())


class TestGroup_save_and_load:
    # noinspection PyBroadException
    class Test_warn_if_another_function_should_be_used:
        @pytest.mark.parametrize('ext, wrong_load, correct_load', _compute_all_mismatched_combos(_load_funcs))
        def test_using_the_wrong_loader_should_always_warn(self, ext, wrong_load, correct_load):
            # The warning should be triggered regardless of whether the file exists or not.
            with pytest.warns(UserWarning, match=f'Use "{correct_load}" for ".{ext}" files.'):
                try:
                    wrong_load(get_unique_path('{}.' + ext))
                except Exception:
                    pass

        @pytest.mark.parametrize('ext, wrong_save, correct_save', _compute_all_mismatched_combos(_save_funcs))
        def test_using_the_wrong_saver_should_always_warn(self, ext, wrong_save, correct_save):
            with pytest.warns(UserWarning, match=f'Use "{correct_save}" for ".{ext}" files.'):
                try:
                    wrong_save('', get_unique_path('{}.' + ext))
                except Exception:
                    pass

    # noinspection PyTypeChecker
    class Test_csv:
        def test_atomicity_interrupt(self):
            path = get_unique_path('{}.csv')
            data = [{'path': path}]

            save_csv.__katalytic_test_atomicity_interrupt__ = True
            save_csv(data, path)
            assert not Path(path).exists()

            # make sure it's still working after the test
            # the atomicity flag is set back to False inside the function
            save_csv(data, path)
            assert load(path) == data

        def test_atomicity_race_condition_error(self):
            path = get_unique_path('{}.csv')
            data = [{'path': path}]

            save_csv.__katalytic_test_atomicity_race_condition__ = True
            assert not Path(path).exists()
            with pytest.raises(FileExistsError):
                save_csv(data, path, exists='error')

            assert load(path) == [{'race': 'condition'}]

            # make sure it's still working after the test
            # the atomicity flag is set back to False inside the function
            delete_file(path)
            assert not Path(path).exists()
            save_csv(data, path, exists='error')
            assert load(path) == data

        def test_atomicity_race_condition_replace(self):
            path = get_unique_path('{}.csv')
            data = [{'path': path}]

            save_csv.__katalytic_test_atomicity_race_condition__ = True
            assert not Path(path).exists()
            save_csv(data, path, exists='replace')
            assert load(path) == data

            # make sure it's still working after the test
            # the atomicity flag is set back to False inside the function
            delete_file(path)
            assert not Path(path).exists()
            save_csv(data, path, exists='replace')
            assert load(path) == data

        def test_atomicity_race_condition_skip(self):
            path = get_unique_path('{}.csv')
            data = [{'path': path}]

            save_csv.__katalytic_test_atomicity_race_condition__ = True
            assert not Path(path).exists()
            save_csv(data, path, exists='skip')
            assert load(path) == [{'race': 'condition'}]

            # make sure it's still working after the test
            # the atomicity flag is set back to False inside the function
            delete_file(path)
            assert not Path(path).exists()
            save_csv(data, path, exists='skip')
            assert load(path) == data

        def test_default(self):
            path = get_unique_path('{}.csv')
            default = [{'a': 1, 'b': 2}]
            assert load_csv(path, default=default) == default

        @pytest.mark.parametrize('mistake', [0, None, {}, True, 'hello'])
        def test_precondition_exists(self, mistake):
            path = get_unique_path('{}.csv')
            with pytest.raises(ValueError):
                save_csv('', path, exists=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_make_dirs(self, mistake):
            path = get_unique_path('{}.csv')
            with pytest.raises(TypeError):
                save_csv('', path, make_dirs=mistake)

        def test_make_dirs_False(self):
            path = get_unique_path('{}/data.csv')
            with pytest.raises(FileNotFoundError):
                save_csv(_create_seq_of_dicts(), path, make_dirs=False)

            path = _setup_file('{}/data')
            with pytest.raises(NotADirectoryError):
                save_csv(_create_seq_of_dicts(), f'{path}/x.csv', make_dirs=False)

        def test_exists_error(self):
            path = _setup_file('{}.csv')
            with pytest.raises(FileExistsError):
                save_csv([], path, exists='error')

        def test_exists_replace(self):
            path = _setup_file('{}.csv')
            data = [{'a': 1, 'b': 2}]
            save_csv(data, path, exists='replace')
            assert load_csv(path) == data

        def test_exists_skip(self):
            path = get_unique_path('{}.csv')
            data_before = [{'a': 1, 'b': 2}]
            data_after = [{'x': 3, 'y': 4}]

            save_csv(data_before, path)
            save_csv(data_after, path, exists='skip')
            assert load_csv(path) == data_before

        @pytest.mark.parametrize('data', [
            _create_seq_of_seq(),
            _create_seq_of_dicts(),
            _create_dict_of_seq(),
        ])
        def test_basic(self, data):
            expected = _create_seq_of_dicts()
            expected[0]['e'] = str(expected[0]['e'])
            expected[2]['c'] = None

            path = _setup_path()
            save_csv(data, path)
            assert load_csv(path) == expected

        def test_mixed_int_with_float(self):
            data = [{'a': 1, 'b': 2.0}, {'a': 3.0, 'b': 4}]
            expected = [{'a': 1.0, 'b': 2}, {'a': 3.0, 'b': 4.0}]

            path = _setup_path()
            save_csv(data, path)
            assert load_csv(path) == expected

        @pytest.mark.parametrize('data', [
            [{'a': 1, 'b': 2}, {'b': 4, 'a': 3}],
            [{'b': 4, 'a': 3}, {'a': 1, 'b': 2}],
        ])
        def test_mixed_column_order(self, data):
            path = _setup_path()
            save_csv(data, path)
            result = load_csv(path)

            assert result == data
            assert dicts_share_key_order(data[0], result[0])
            assert dicts_share_key_order(data[0], result[1])
            assert not dicts_share_key_order(data[0], data[1])

    class Test_json:
        def test_atomicity_interrupt(self):
            path = get_unique_path('{}.json')
            data = [{'path': path}]

            try:
                save_json.__katalytic_test_atomicity_interrupt__ = True
                save_json(data, path)
                assert not Path(path).exists()

                # make sure it's still working after the test
                # the atomicity flag is set back to False inside the function
                save_json(data, path)
                assert load_json(path) == data
            except:
                raise

        def test_atomicity_race_condition_error(self):
            path = get_unique_path('{}.json')
            data = [{'path': path}]

            try:
                save_json.__katalytic_test_atomicity_race_condition__ = True
                assert not Path(path).exists()
                with pytest.raises(FileExistsError):
                    save_json(data, path, exists='error')

                assert load(path) == [{'race': 'condition'}]

                # make sure it's still working after the test
                # the atomicity flag is set back to False inside the function
                delete_file(path)
                assert not Path(path).exists()
                save_json(data, path, exists='error')
                assert load(path) == data
            except:
                raise

        def test_atomicity_race_condition_replace(self):
            path = get_unique_path('{}.json')
            data = [{'path': path}]

            try:
                save_json.__katalytic_test_atomicity_race_condition__ = True
                assert not Path(path).exists()
                save_json(data, path, exists='replace')
                assert load(path) == data

                # make sure it's still working after the test
                # the atomicity flag is set back to False inside the function
                delete_file(path)
                assert not Path(path).exists()
                save_json(data, path, exists='replace')
                assert load(path) == data
            except:
                raise

        def test_atomicity_race_condition_skip(self):
            path = get_unique_path('{}.json')
            data = [{'path': path}]

            try:
                save_json.__katalytic_test_atomicity_race_condition__ = True
                assert not Path(path).exists()
                save_json(data, path, exists='skip')
                assert load(path) == [{'race': 'condition'}]

                # make sure it's still working after the test
                # the atomicity flag is set back to False inside the function
                delete_file(path)
                assert not Path(path).exists()
                save_json(data, path, exists='skip')
                assert load(path) == data
            except:
                raise

        def test_default(self):
            path = get_unique_path('{}.json')
            default = [{'a': 1, 'b': 2}]
            assert load_json(path, default=default) == default

        @pytest.mark.parametrize('mistake', [0, None, {}, True, 'hello'])
        def test_precondition_exists(self, mistake):
            path = get_unique_path('{}.json')
            with pytest.raises(ValueError):
                save_json([], path, exists=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_make_dirs(self, mistake):
            path = get_unique_path('{}.json')
            with pytest.raises(TypeError):
                save_json([], path, make_dirs=mistake)

        def test_make_dirs_False(self):
            path = get_unique_path('{}/data.json')
            with pytest.raises(FileNotFoundError):
                save_json(_create_seq_of_dicts(), path, make_dirs=False)

            path = _setup_file('{}/data')
            with pytest.raises(NotADirectoryError):
                save_json(_create_seq_of_dicts(), f'{path}/x.json', make_dirs=False)

        def test_exists_error(self):
            path = _setup_file('{}.json')
            with pytest.raises(FileExistsError):
                save_json([], path, exists='error')

        def test_exists_replace(self):
            path = _setup_file('{}.json')
            data = [{'a': 1, 'b': 2}]
            save_json(data, path, exists='replace')
            assert load_json(path) == data

        def test_exists_skip(self):
            path = get_unique_path('{}.json')
            data_before = [{'a': 1, 'b': 2}]
            data_after = [{'x': 3, 'y': 4}]

            save_json(data_before, path)
            save_json(data_after, path, exists='skip')
            assert load_json(path) == data_before

        def test_indent_0(self):
            data = self.create_json_data()
            path = _setup_path()
            save_json(data, path, indent=0)
            assert load_json(path) == data

        def test_indent_positive(self):
            data = self.create_json_data()
            path = _setup_path()
            save_json(data, path, indent=4)
            assert load_json(path) == data

        def test_indent_negative(self):
            data = self.create_json_data()
            path = _setup_path()
            with pytest.raises(ValueError):
                save_json(data, path, indent=-1)

        def test_indent_as_float(self):
            data = self.create_json_data()
            path = _setup_path()

            save_json(data, path, indent=1.0)
            assert load_json(path) == data

        @pytest.mark.parametrize('mistake', all_types_besides('int'))
        def test_indent_of_wrong_type(self, mistake):
            data = self.create_json_data()
            path = _setup_path()
            with pytest.raises(TypeError):
                save_json(data, path, indent=mistake)

        def test_save_sort_keys(self):
            data = self.create_json_data()
            path = _setup_path()
            save_json(data, path, sort_keys=True)
            assert self.are_keys_sorted_recursive(load_json(path))

        def test_save_sort_keys_False(self):
            data = self.create_json_data()
            path = _setup_path()
            save_json(data, path, sort_keys=False)
            assert not self.are_keys_sorted_recursive(load_json(path))

        def are_keys_sorted_recursive(self, data):
            if isinstance(data, dict):
                if list(data.keys()) != sorted(data.keys()):
                    return False

                return all(self.are_keys_sorted_recursive(v) for v in data.values())
            elif is_iterable(data) and not isinstance(data, str):
                return all(self.are_keys_sorted_recursive(item) for item in data)
            else:
                return True

        @staticmethod
        def create_json_data():
            """The keys are shuffled on purpose."""
            return {'e': 0, 'a': 1, 'b': 2.0, 'c': 3.5, 'd': [{'y': None, 'f': True, 'g': {'x': 'i', 'j': [False]}}]}

    class Test_text:
        def test_atomicity_interrupt(self):
            path = get_unique_path('{}.txt')

            try:
                save_text.__katalytic_test_atomicity_interrupt__ = True
                save_text(path, path)
                assert not Path(path).exists()

                # make sure it's still working after the test
                # the atomicity flag is set back to False inside the function
                save_text(path, path)
                assert load(path) == path
            except:
                raise

        def test_atomicity_race_condition_error(self):
            path = get_unique_path('{}.txt')
            data = f'path = {path!r}'

            try:
                save_text.__katalytic_test_atomicity_race_condition__ = True
                assert not Path(path).exists()
                with pytest.raises(FileExistsError):
                    save_text(data, path, exists='error')

                assert load(path) == 'race condition'

                # make sure it's still working after the test
                # the atomicity flag is set back to False inside the function
                delete_file(path)
                assert not Path(path).exists()
                save_text(data, path, exists='error')
                assert load(path) == data
            except:
                raise

        def test_atomicity_race_condition_replace(self):
            path = get_unique_path('{}.txt')
            data = f'path = {path!r}'

            try:
                save_text.__katalytic_test_atomicity_race_condition__ = True
                assert not Path(path).exists()
                save_text(data, path, exists='replace')
                assert load(path) == data

                # make sure it's still working after the test
                # the atomicity flag is set back to False inside the function
                delete_file(path)
                assert not Path(path).exists()
                save_text(data, path, exists='replace')
                assert load(path) == data
            except:
                raise

        def test_atomicity_race_condition_skip(self):
            path = get_unique_path('{}.txt')
            data = f'path = {path!r}'

            try:
                save_text.__katalytic_test_atomicity_race_condition__ = True
                assert not Path(path).exists()
                save_text(data, path, exists='skip')
                assert load(path) == 'race condition'

                # make sure it's still working after the test
                # the atomicity flag is set back to False inside the function
                delete_file(path)
                assert not Path(path).exists()
                save_text(data, path, exists='skip')
                assert load(path) == data
            except:
                raise

        def test_default(self):
            path = get_unique_path('{}.txt')
            assert load_text(path, default='hello') == 'hello'

        @pytest.mark.parametrize('mistake', [0, None, {}, True, 'hello'])
        def test_precondition_exists(self, mistake):
            path = get_unique_path('{}.txt')
            with pytest.raises(ValueError):
                save_text('', path, exists=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_make_dirs(self, mistake):
            path = get_unique_path('{}.txt')
            with pytest.raises(TypeError):
                save_text('', path, make_dirs=mistake)

        def test_make_dirs_False(self):
            path = get_unique_path('{}/data.txt')
            with pytest.raises(FileNotFoundError):
                save_text('hello', path, make_dirs=False)

            path = _setup_file('{}/data')
            with pytest.raises(NotADirectoryError):
                save_text('hello', f'{path}/x.txt', make_dirs=False)

        def test_exists_error(self):
            path = _setup_file('{}.txt')
            with pytest.raises(FileExistsError):
                save_text('data', path, exists='error')

        def test_exists_replace(self):
            path = _setup_file('{}.txt')
            save_text('data', path, exists='replace')
            assert load_text(path) == 'data'

        def test_exists_skip(self):
            path = get_unique_path('{}.txt')
            save_text('before', path)
            save_text('after', path, exists='skip')
            assert load_text(path) == 'before'

        def test_basic(self):
            data = 'some text\nwith newlines'
            path = _setup_path()
            save_text(data, path)
            assert load_text(path) == data

    # noinspection PyTypeChecker
    class Test_universal_load_and_save:
        @pytest.mark.parametrize('pattern, default', [
            ('{}.csv', _create_seq_of_dicts()),
            ('{}.json', _create_dict_of_seq()),
            ('{}.txt', 'hello'),
        ])
        def test_default(self, pattern, default):
            path = get_unique_path(pattern)
            assert load(path, default=default) == default

        def test_unknown(self):
            with pytest.raises(RuntimeError):
                load('data.unknown')

            with pytest.raises(RuntimeError):
                save('', 'data.unknown')

        def test_txt(self):
            data = 'Hello, World!\nThis is not a drill'
            path = _setup_path('{}.txt')
            save(data, path)

            assert load(path) == load_text(path) == data

        def test_csv(self):
            data = _create_seq_of_dicts()
            path = _setup_path('{}.CSV')
            save(data, path)

            expected_csv = copy.deepcopy(data)
            expected_csv[0]['e'] = str(expected_csv[0]['e'])
            expected_csv[2]['c'] = None

            expected_text = '\n'.join([
                '"a","b","c","d","e"',
                '"1","x","True","-1.5","True"',
                '"2","","False","0.0","z"',
                '"","y","","","w"',
            ])

            assert load(path) == expected_csv

            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                assert load_text(path).strip() == expected_text

        def test_json(self):
            data = {'a': 1, 'b': [2, True, None, {}]}
            path = _setup_path('{}.json')
            save(data, path)

            expected_text = '\n'.join([
                '{',
                '    "a": 1,',
                '    "b": [',
                '        2,',
                '        true,',
                '        null,',
                '        {}',
                '    ]',
                '}',
            ])

            assert load(path) == data

            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                assert load_text(path).strip() == expected_text


class TestGroup_copy:
    # noinspection PyTypeChecker
    class Test_copy_dir:
        def test_dest_is_dir(self):
            src = Path(_setup_tree('{}_src'))

            dest = Path(get_unique_path('{}_dest'))
            copy_dir(src, dest)
            _check_tree(src, dest/src.name)

            dest_2 = Path(get_unique_path('{}_dest_2'))
            copy_dir(src, dest_2/src.name)
            _check_tree(src, dest_2/src.name)

        def test_dest_is_file(self):
            with pytest.raises(NotADirectoryError):
                src = _setup_dir()
                dest = _setup_file()
                copy_dir(src, dest)

        def test_dir_exists_error(self):
            with pytest.raises(FileExistsError):
                src = _setup_tree()
                dest = _setup_tree()
                copy_dir(src, dest, dir_exists='error', file_exists='skip')

        def test_dir_exists_merge(self):
            src, src_file = _setup_dir_and_file()
            _setup_tree(src + '/{}')

            dest = _setup_tree()
            copy_dir(src, dest, dir_exists='merge', file_exists='skip')

            src_dirs = get_dirs(src, prefix=False, recursive=True)
            dest_dirs = get_dirs(dest, prefix=False, recursive=True)
            assert set(src_dirs).difference(dest_dirs) == set()
            assert set(dest_dirs).difference(src_dirs) != set()

        def test_dir_exists_replace(self):
            src, src_file = _setup_dir_and_file()
            dest = _setup_tree()
            copy_dir(src, dest, dir_exists='replace', file_exists='skip')

            src_tree = get_all(src, prefix=False, recursive=True)
            dest_tree = get_all(dest, prefix=False, recursive=True)
            assert src_tree == dest_tree

            src_tree = set(get_all(src, recursive=True))
            dest_tree = set(get_all(dest, recursive=True))
            pairs = {(fn, fn.replace(src, dest)) for fn in src_tree | dest_tree}
            pairs = {(Path(s), Path(d)) for s, d in pairs if Path(s).is_file()}
            assert all((s.read_text() == d.read_text()) for s, d in pairs)

        def test_dir_exists_skip(self):
            src, src_file = _setup_dir_and_file()
            dest, dest_file = _setup_dir_and_file()
            make_dir(f'{dest}/subdir')

            copy_dir(src, dest, dir_exists='skip', file_exists='replace')
            assert Path(dest_file).read_text() != Path(src_file).read_text()
            assert Path(f'{dest}/subdir').is_dir()

        def test_file_exists_error(self):
            with pytest.raises(FileExistsError):
                src = _setup_tree()
                dest = _setup_tree()
                copy_dir(src, dest, dir_exists='merge', file_exists='error')

        def test_file_exists_replace(self):
            src, src_file = _setup_dir_and_file()
            dest, dest_file = _setup_dir_and_file()

            copy_dir(src, dest, dir_exists='merge', file_exists='replace')
            assert Path(dest_file).read_text() == Path(src_file).read_text()

        def test_file_exists_skip(self):
            src, src_file = _setup_dir_and_file()
            dest, dest_file = _setup_dir_and_file()

            original = Path(dest_file).read_text()
            copy_dir(src, dest, dir_exists='merge', file_exists='skip')
            assert Path(dest_file).read_text() != Path(src_file).read_text()
            assert Path(dest_file).read_text() == original

        def test_make_dirs_false(self):
            with pytest.raises(FileNotFoundError):
                with tempfile.TemporaryDirectory() as d:
                    src = get_unique_path(d + '/{}_src')
                    dest = get_unique_path(d + '/{}_dest')
                    make_dir(src)
                    copy_dir(src, dest, make_dirs=False)

        def test_make_dirs_true(self):
            with tempfile.TemporaryDirectory() as d:
                src = _setup_tree()
                src_name = Path(src).name

                dest = get_unique_path(d + '/{}_dest')
                copy_dir(src, dest, make_dirs=True)
                _check_tree(src, f'{dest}/{src_name}')

                dest_2 = get_unique_path(d + '/{}_dest_2')
                copy_dir(src, f'{dest_2}/{src_name}', make_dirs=True)
                _check_tree(src, f'{dest_2}/{src_name}')

        def test_same_path(self):
            with pytest.raises(ValueError):
                src = _setup_dir()
                copy_dir(src, src)

        def test_src_is_file(self):
            with pytest.raises(NotADirectoryError):
                src = _setup_file()
                dest = get_unique_path('{}_new')
                copy_dir(src, dest)

        def test_src_is_missing(self):
            with pytest.raises(FileNotFoundError):
                src = get_unique_path()
                copy_dir(src, f'{src}_copy')

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_dest(self, mistake):
            with pytest.raises(TypeError):
                copy_dir('src', mistake)

        @pytest.mark.parametrize('mistake', [0, None, {}, True, 'hello'])
        def test_precondition_dir_exists(self, mistake):
            with pytest.raises(ValueError):
                copy_dir('src', 'dest', dir_exists=mistake)

        @pytest.mark.parametrize('mistake', [0, None, {}, 'hello'])
        def test_precondition_file_exists(self, mistake):
            with pytest.raises(ValueError):
                copy_dir('src', 'dest', file_exists=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_make_dirs(self, mistake):
            with pytest.raises(TypeError):
                src, dest = _setup_dir(n=2)
                copy_dir(src, dest, make_dirs=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_src(self, mistake):
            with pytest.raises(TypeError):
                copy_dir(mistake, 'dest')

    # noinspection PyTypeChecker
    class Test_copy_file:
        def test_atomicity_interrupt(self):
            src = _setup_file('{}.txt')
            dest = get_unique_path('{}.txt')

            copy_file.__katalytic_test_atomicity_interrupt__ = True
            copy_file(src, dest)
            assert not Path(dest).exists()

            # make sure it's still working after the test
            # the atomicity flag is set back to False inside the function
            copy_file(src, dest)
            assert load(dest) == src

        def test_atomicity_race_condition_error(self):
            src = _setup_file('{}.txt')
            dest = get_unique_path('{}.txt')

            copy_file.__katalytic_test_atomicity_race_condition__ = True
            assert not Path(dest).exists()
            with pytest.raises(FileExistsError):
                copy_file(src, dest, exists='error')

            assert load(dest) == 'race condition'

            # make sure it's still working after the test
            # the atomicity flag is set back to False inside the function
            delete_file(dest)
            assert not Path(dest).exists()
            copy_file(src, dest, exists='error')
            assert load(dest) == src

        def test_atomicity_race_condition_replace(self):
            src = _setup_file('{}.txt')
            dest = get_unique_path('{}.txt')

            copy_file.__katalytic_test_atomicity_race_condition__ = True
            assert not Path(dest).exists()
            copy_file(src, dest, exists='replace')
            assert load(dest) == src

            # make sure it's still working after the test
            # the atomicity flag is set back to False inside the function
            delete_file(dest)
            assert not Path(dest).exists()
            copy_file(src, dest, exists='replace')
            assert load(dest) == src

        def test_atomicity_race_condition_skip(self):
            src = _setup_file('{}.txt')
            dest = get_unique_path('{}.txt')

            copy_file.__katalytic_test_atomicity_race_condition__ = True
            assert not Path(dest).exists()
            copy_file(src, dest, exists='skip')
            assert load(dest) == 'race condition'

            # make sure it's still working after the test
            # the atomicity flag is set back to False inside the function
            delete_file(dest)
            assert not Path(dest).exists()
            copy_file(src, dest, exists='skip')
            assert load(dest) == src

        def test_dir(self):
            with pytest.raises(IsADirectoryError):
                src = _setup_dir()
                dest = get_unique_path()
                copy_file(src, dest)

        def test_exists_error(self):
            with pytest.raises(FileExistsError):
                src, dest = _setup_file(n=2)
                copy_file(src, dest, exists='error')

        def test_exists_replace(self):
            src, dest = _setup_file(n=2)
            copy_file(src, dest, exists='replace')

            text = Path(dest).read_text()
            assert text == Path(src).read_text()
            assert text != dest

        def test_exists_skip(self):
            src, dest = _setup_file(n=2)
            copy_file(src, dest, exists='skip')

            text = Path(dest).read_text()
            assert text != Path(src).read_text()
            assert text == dest

        def test_missing(self):
            dest = _setup_file()
            src = get_unique_path()
            with pytest.raises(FileNotFoundError):
                copy_file(src, dest)

        def test_ends_with_slash(self):
            src = _setup_file()
            root = get_unique_path()
            dest = root + '/1/2/3/'
            copy_file(src, dest)

        def test_make_dirs_false(self):
            with pytest.raises(FileNotFoundError):
                src = _setup_file()
                dest = get_unique_path('{}/a/b/c/')
                copy_file(src, dest, make_dirs=False)

        def test_make_dirs_true(self):
            src = _setup_file()
            dest = _setup_path()
            copy_file(src, dest, make_dirs=True)
            assert Path(dest).read_text() == src

        def test_same_path(self):
            with pytest.raises(ValueError):
                src = _setup_file()
                copy_file(src, src)

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_src(self, mistake):
            with pytest.raises(TypeError):
                copy_file(mistake, 'a')

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_dest(self, mistake):
            with pytest.raises(TypeError):
                copy_file('a', mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_make_dirs(self, mistake):
            with pytest.raises(TypeError):
                src, dest = _setup_file(n=2)
                copy_file(src, dest, make_dirs=mistake)

        @pytest.mark.parametrize('mistake', [0, None, {}, 'hello'])
        def test_precondition_exists(self, mistake):
            src, dest = _setup_file(n=2)
            with pytest.raises(ValueError):
                copy_file(src, dest, exists=mistake)


class TestGroup_delete:
    # noinspection PyTypeChecker
    class Test_delete_dir:
        def test_file(self):
            with pytest.raises(NotADirectoryError):
                path = _setup_file()
                delete_dir(path)

        def test_empty_dir(self):
            path = _setup_dir()
            delete_dir(path)
            assert not Path(path).exists()

        def test_missing_ok_false(self):
            with pytest.raises(FileNotFoundError):
                path = get_unique_path()
                delete_dir(path, missing_ok=False)

        def test_missing_ok_true(self):
            path = get_unique_path()
            delete_dir(path, missing_ok=True)
            assert not Path(path).exists()

        def test_non_empty_dir_false(self):
            path, _ = _setup_dir_and_file()
            delete_dir(path, non_empty_dir=False)
            assert Path(path).is_dir()

        def test_non_empty_dir_error(self):
            with pytest.raises(OSError):
                path, _ = _setup_dir_and_file()
                delete_dir(path, non_empty_dir='error')

        def test_non_empty_dir_true(self):
            path, _ = _setup_dir_and_file()
            delete_dir(path, non_empty_dir=True)
            assert not Path(path).exists()

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_missing_ok(self, mistake):
            with pytest.raises(TypeError):
                delete_dir(get_unique_path(), missing_ok=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_non_empty_dir(self, mistake):
            with pytest.raises(ValueError):
                delete_dir(get_unique_path(), non_empty_dir=mistake)

        @pytest.mark.parametrize('mistake', [None, '', '.'])
        def test_precondition_path_current_dir(self, mistake):
            with pytest.raises(ValueError):
                delete_dir(mistake)

        @pytest.mark.parametrize('mistake', all_types_besides(['none', 'str', 'path']))
        def test_precondition_path_type(self, mistake):
            with pytest.raises(TypeError):
                delete_dir(mistake)

    # noinspection PyTypeChecker
    class Test_delete_file:
        def test_dir(self):
            with pytest.raises(IsADirectoryError):
                path = _setup_dir()
                delete_file(path)

        def test_file(self):
            path = _setup_file()
            delete_file(path)
            assert not Path(path).exists()

        def test_missing_ok_false(self):
            with pytest.raises(FileNotFoundError):
                path = get_unique_path()
                delete_file(path, missing_ok=False)

        def test_missing_ok_true(self):
            path = get_unique_path()
            delete_file(path, missing_ok=True)
            assert not Path(path).exists()

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_path(self, mistake):
            with pytest.raises(TypeError):
                delete_file(mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_missing_ok(self, mistake):
            with pytest.raises(TypeError):
                delete_file(get_unique_path(), missing_ok=mistake)


class TestGroup_get:
    class Test_get_all:
        def test_prefix_false(self):
            path = _setup_tree()
            expected = ['1.txt', 'a', 'a/2.txt', 'b', 'b/3.txt', 'b/c', 'b/c/4.txt']
            assert get_all(path, prefix=False, recursive=True) == expected

        def test_prefix_true(self):
            path = _setup_tree()
            expected = [
                f'{path}/1.txt',
                f'{path}/a',
                f'{path}/a/2.txt',
                f'{path}/b',
                f'{path}/b/3.txt',
                f'{path}/b/c',
                f'{path}/b/c/4.txt'
            ]
            assert get_all(path, prefix=True, recursive=True) == expected

        def test_path_is_file(self):
            with pytest.raises(NotADirectoryError):
                path = _setup_file()
                get_all(path)

        def test_path_is_missing(self):
            with pytest.raises(FileNotFoundError):
                get_all(get_unique_path())

        def test_iter(self):
            r = get_all('', iter_=True)
            assert is_iterator(r)

        def test_glob(self):
            path = _setup_tree_2()
            assert get_all(f'{path}/*/a/*') == [f'{path}/a/a/6.py', f'{path}/b/a/5.py']

        def test_glob_recursive(self):
            path = _setup_tree_2()
            expected = sorted(map(str, Path(path).glob('**/a/*')))
            assert get_all(f'{path}/**/a/*') == expected
            assert get_all(f'{path}/*/a/*', recursive=True) == expected

            expected = sorted(map(str, Path(path).glob('**/a/**')))
            assert get_all(f'{path}/**/a/**') == expected
            assert get_all(f'{path}/**/a/**', recursive=True) == expected

        def test_only_dirs(self):
            path = _setup_tree()
            assert _get_all(path, only_dirs=True, prefix=False) == ['a', 'b']

        def test_only_files(self):
            path = _setup_tree()
            assert _get_all(path, only_files=True, prefix=False) == ['1.txt']

        def test_only_conflict(self):
            with pytest.raises(ValueError):
                _get_all('', only_files=True, only_dirs=True)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_glob(self, mistake):
            with pytest.raises(TypeError):
                get_all('', glob=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_iter_(self, mistake):
            with pytest.raises(TypeError):
                get_all('', iter_=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_only_dirs(self, mistake):
            with pytest.raises(TypeError):
                _get_all('', only_dirs=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_only_files(self, mistake):
            with pytest.raises(TypeError):
                _get_all('', only_files=mistake)

        @pytest.mark.parametrize('mistake', [0, None, {}, True])
        def test_precondition_path(self, mistake):
            with pytest.raises(TypeError):
                get_all(mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_prefix(self, mistake):
            with pytest.raises(TypeError):
                get_all('', prefix=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_recursive(self, mistake):
            with pytest.raises(TypeError):
                get_all('', recursive=mistake)

        def test_recursive(self):
            path = _setup_tree()
            assert get_all(path, recursive=True) == [
                f'{path}/1.txt',
                f'{path}/a',
                f'{path}/a/2.txt',
                f'{path}/b',
                f'{path}/b/3.txt',
                f'{path}/b/c',
                f'{path}/b/c/4.txt',
            ]

        def test_sorted(self):
            path = _setup_tree()
            _ = _setup_tree(path + '/b/{}')
            assert get_all(path, iter_=False, recursive=True) == [
                f'{path}/1.txt',
                f'{path}/a',
                f'{path}/a/2.txt',
                f'{path}/b',
                f'{path}/b/1',
                f'{path}/b/1/1.txt',
                f'{path}/b/1/a',
                f'{path}/b/1/a/2.txt',
                f'{path}/b/1/b',
                f'{path}/b/1/b/3.txt',
                f'{path}/b/1/b/c',
                f'{path}/b/1/b/c/4.txt',
                f'{path}/b/3.txt',
                f'{path}/b/c',
                f'{path}/b/c/4.txt',
            ]

        def test_returns_same_type_as_input(self):
            path = _setup_tree()
            result = get_all(path, recursive=True)
            assert all(isinstance(p, str) for p in result)

            path = Path(_setup_tree())
            result = get_all(path, recursive=True)
            assert all(isinstance(p, Path) for p in result)

    class Test_get_dirs:
        def test_prefix_false(self):
            path = _setup_tree()
            assert get_dirs(path, prefix=False, recursive=True) == ['a', 'b', 'b/c']

        def test_prefix_true(self):
            path = _setup_tree()
            expected = [f'{path}/a', f'{path}/b', f'{path}/b/c']
            assert get_dirs(path, prefix=True, recursive=True) == expected

        def test_path_is_file(self):
            with pytest.raises(NotADirectoryError):
                path = _setup_file()
                get_dirs(path)

        def test_path_is_missing(self):
            with pytest.raises(FileNotFoundError):
                get_dirs(get_unique_path())

        def test_iter(self):
            r = get_dirs('', iter_=True)
            assert is_iterator(r)

        def test_glob(self):
            path = _setup_tree_2()
            assert get_dirs(f'{path}/*/a') == [f'{path}/a/a', f'{path}/b/a']

        def test_glob_recursive(self):
            path = _setup_tree_2()
            expected = [f'{path}/a', f'{path}/a/a', f'{path}/b/a', f'{path}/b/c/a']
            assert get_dirs(f'{path}/**/a') == expected
            assert get_dirs(f'{path}/*/a', recursive=True) == expected

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_prefix(self, mistake):
            with pytest.raises(TypeError):
                get_dirs('', prefix=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_glob(self, mistake):
            with pytest.raises(TypeError):
                get_dirs('', glob=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_iter_(self, mistake):
            with pytest.raises(TypeError):
                get_dirs('', iter_=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_path(self, mistake):
            with pytest.raises(TypeError):
                get_dirs(mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_recursive(self, mistake):
            with pytest.raises(TypeError):
                get_dirs('', recursive=mistake)

        def test_recursive(self):
            path = _setup_tree()
            assert get_dirs(path, recursive=True) == [
                f'{path}/a',
                f'{path}/b',
                f'{path}/b/c',
            ]

        def test_sorted(self):
            path = _setup_tree()
            _ = _setup_tree(path + '/b/{}')
            assert get_dirs(path, iter_=False, recursive=True) == [
                f'{path}/a',
                f'{path}/b',
                f'{path}/b/1',
                f'{path}/b/1/a',
                f'{path}/b/1/b',
                f'{path}/b/1/b/c',
                f'{path}/b/c',
            ]

        def test_returns_same_type_as_input(self):
            path = _setup_tree()
            result = get_dirs(path, recursive=True)
            assert all(isinstance(p, str) for p in result)

            path = Path(_setup_tree())
            result = get_dirs(path, recursive=True)
            assert all(isinstance(p, Path) for p in result)

    class Test_get_files:
        def test_prefix_false(self):
            path = _setup_tree()
            expected = ['1.txt', 'a/2.txt', 'b/3.txt', 'b/c/4.txt']
            assert get_files(path, prefix=False, recursive=True) == expected

        def test_prefix_true(self):
            path = _setup_tree()
            expected = [
                f'{path}/1.txt',
                f'{path}/a/2.txt',
                f'{path}/b/3.txt',
                f'{path}/b/c/4.txt'
            ]
            assert get_files(path, prefix=True, recursive=True) == expected

        def test_path_is_file(self):
            with pytest.raises(NotADirectoryError):
                path = _setup_file()
                get_files(path)

        def test_path_is_missing(self):
            with pytest.raises(FileNotFoundError):
                get_files(get_unique_path())

        def test_iter(self):
            r = get_files('', iter_=True)
            assert is_iterator(r)

        def test_glob(self):
            path = _setup_tree_2()
            assert get_files(f'{path}/*.py') == [f'{path}/1.py']

        def test_glob_recursive(self):
            path = _setup_tree_2()
            expected = [f'{path}/1.py', f'{path}/a/2.py', f'{path}/a/a/6.py', f'{path}/b/a/5.py']
            assert get_files(f'{path}/**/*.py') == expected

        def test_only_conflict(self):
            with pytest.raises(ValueError):
                _get_all('', only_files=True, only_dirs=True)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_prefix(self, mistake):
            with pytest.raises(TypeError):
                get_files('', prefix=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_glob(self, mistake):
            with pytest.raises(TypeError):
                get_files('', glob=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_iter_(self, mistake):
            with pytest.raises(TypeError):
                get_files('', iter_=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_path(self, mistake):
            with pytest.raises(TypeError):
                get_files(mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_recursive(self, mistake):
            with pytest.raises(TypeError):
                get_files('', recursive=mistake)

        def test_recursive(self):
            path = _setup_tree()
            assert get_files(path, recursive=True) == [
                f'{path}/1.txt',
                f'{path}/a/2.txt',
                f'{path}/b/3.txt',
                f'{path}/b/c/4.txt',
            ]

        def test_sorted(self):
            path = _setup_tree()
            _ = _setup_tree(path + '/b/{}')
            assert get_files(path, iter_=False, recursive=True) == [
                f'{path}/1.txt',
                f'{path}/a/2.txt',
                f'{path}/b/1/1.txt',
                f'{path}/b/1/a/2.txt',
                f'{path}/b/1/b/3.txt',
                f'{path}/b/1/b/c/4.txt',
                f'{path}/b/3.txt',
                f'{path}/b/c/4.txt',
            ]

        def test_returns_same_type_as_input(self):
            path = _setup_tree()
            result = get_files(path, recursive=True)
            assert all(isinstance(p, str) for p in result)

            path = Path(_setup_tree())
            result = get_files(path, recursive=True)
            assert all(isinstance(p, Path) for p in result)


class TestGroup_misc_dir:
    # noinspection PyTypeChecker
    class Test_clear_dir:
        def test_clear_empty(self):
            path = _setup_dir()
            clear_dir(path)
            assert list(Path(path).iterdir()) == []

        def test_clear_non_empty(self):
            path, _ = _setup_dir_and_file()
            clear_dir(path)
            assert list(Path(path).iterdir()) == []

        def test_create_missing_false(self):
            d = get_unique_path()
            shutil.rmtree(d, ignore_errors=True)
            clear_dir(d, create_missing=False)
            assert not Path(d).exists()

        def test_create_missing_true(self):
            path = get_unique_path()
            clear_dir(path, create_missing=True)
            assert is_dir_empty(path)

        def test_is_file(self):
            with pytest.raises(NotADirectoryError):
                file = _setup_file()
                clear_dir(file)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_create_missing(self, mistake):
            path = get_unique_path()
            with pytest.raises(TypeError):
                clear_dir(path, create_missing=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_path(self, mistake):
            with pytest.raises(TypeError):
                clear_dir(mistake)

    # noinspection PyTypeChecker
    class Test_is_dir_empty:
        def test_empty(self):
            path = _setup_dir()
            assert is_dir_empty(path)

        def test_is_file(self):
            with pytest.raises(NotADirectoryError):
                path = _setup_file()
                is_dir_empty(path)

        def test_missing_error(self):
            with pytest.raises(FileNotFoundError):
                is_dir_empty(get_unique_path(), missing='error')

        def test_missing_false(self):
            path = get_unique_path()
            assert is_dir_empty(path, missing=False) is False

        def test_missing_true(self):
            path = get_unique_path()
            assert is_dir_empty(path, missing=True) is True

        def test_non_empty(self):
            path, _ = _setup_dir_and_file()
            assert not is_dir_empty(path)

        @pytest.mark.parametrize('mistake', [[], '', 0, None])
        def test_precondition_missing(self, mistake):
            with pytest.raises(ValueError):
                path = get_unique_path()
                is_dir_empty(path, missing=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_path(self, mistake):
            with pytest.raises(TypeError):
                is_dir_empty(mistake)

    # noinspection PyTypeChecker
    class Test_mkdir:
        def test_create(self):
            path = _setup_dir()
            assert Path(path).is_dir()

            path = _setup_dir('a/{}/b')
            assert Path(path).is_dir()

        def test_exists_error(self):
            with pytest.raises(FileExistsError):
                path = _setup_dir()
                make_dir(path, exists_ok=False)

        def test_exists_ok(self):
            path = _setup_dir()
            make_dir(path, exists_ok=True)
            assert Path(path).is_dir()

        def test_idempotency(self):
            path = _setup_dir()
            make_dir(path)
            assert Path(path).is_dir()

        def test_is_file(self):
            with pytest.raises(NotADirectoryError):
                path = _setup_file()
                make_dir(path)

        def test_no_parents(self):
            with pytest.raises(FileNotFoundError):
                d = get_unique_path('parent/{}/child')
                make_dir(d, create_parents=False)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_create_parents(self, mistake):
            with pytest.raises(TypeError):
                make_dir(get_unique_path(), create_parents=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_exists_ok(self, mistake):
            with pytest.raises(TypeError):
                make_dir(get_unique_path(), exists_ok=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_path(self, mistake):
            with pytest.raises(TypeError):
                make_dir(mistake)


class TestGroup_misc_file:
    # noinspection PyTypeChecker
    class Test_is_file_empty:
        def test_is_dir(self):
            with pytest.raises(IsADirectoryError):
                is_file_empty(_setup_dir())

        def test_missing_error(self):
            with pytest.raises(FileNotFoundError):
                is_file_empty(get_unique_path(), missing='error')

        def test_missing_false(self):
            path = get_unique_path()
            assert is_file_empty(path, missing=False) is False

        def test_missing_true(self):
            path = get_unique_path()
            assert is_file_empty(path, missing=True) is True

        def test_empty(self):
            path = get_unique_path()
            Path(path).touch()
            assert is_file_empty(path)

        def test_non_empty(self):
            path = get_unique_path()
            Path(path).write_text('hello')
            assert not is_file_empty(path)

        @pytest.mark.parametrize('mistake', [[], '', 0, None])
        def test_precondition_missing(self, mistake):
            with pytest.raises(ValueError):
                path = get_unique_path()
                is_file_empty(path, missing=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_path(self, mistake):
            with pytest.raises(TypeError):
                is_file_empty(mistake)


class TestGroup_move:
    # noinspection PyTypeChecker
    class Test_move_dir:
        def test_dest_is_dir(self):
            src = Path(_setup_tree('{}_src'))
            src_copy = Path(get_unique_path('{}_src_copy'))
            copy_dir(src, src_copy)

            dest = get_unique_path('{}_dest')
            move_dir(src, dest)
            _check_tree(src_copy, dest)
            assert not src.exists()

            src = Path(_setup_tree('{}_src'))
            src_copy = Path(get_unique_path('{}_src_copy'))
            copy_dir(src, src_copy)

            dest = Path(get_unique_path('{}_dest'))
            move_dir(src, dest/src.name)
            _check_tree(src_copy/src.name, dest/src.name)
            assert not src.exists()

        def test_dest_is_file(self):
            with pytest.raises(NotADirectoryError):
                src = _setup_dir()
                dest = _setup_file()
                move_dir(src, dest)

        def test_dir_exists_error(self):
            with pytest.raises(FileExistsError):
                src = _setup_tree()
                dest = _setup_tree()
                move_dir(src, dest, dir_exists='error', file_exists='skip')

        def test_dir_exists_merge(self):
            src, src_file = _setup_dir_and_file()
            _setup_tree(src + '/{}')
            src_dirs = get_dirs(src, prefix=False, recursive=True)

            dest = _setup_tree()
            move_dir(src, dest, dir_exists='merge', file_exists='skip')

            dest_dirs = get_dirs(dest, prefix=False, recursive=True)
            assert set(src_dirs).difference(dest_dirs) == set()
            assert set(dest_dirs).difference(src_dirs) != set()

        def test_dir_exists_replace(self):
            src, src_file = _setup_dir_and_file()
            src_tree = get_all(src, prefix=False, recursive=True)
            src_tree_2 = set(get_all(src, recursive=True))

            dest = _setup_tree()
            move_dir(src, dest, dir_exists='replace', file_exists='skip')

            dest_tree = get_all(dest, prefix=False, recursive=True)
            assert src_tree == dest_tree

            dest_tree = set(get_all(dest, recursive=True))
            pairs = {(fn, fn.replace(src, dest)) for fn in src_tree_2 | dest_tree}
            pairs = {(Path(s), Path(d)) for s, d in pairs if Path(s).is_file()}
            assert all((s.read_text() == d.read_text()) for s, d in pairs)

        def test_dir_exists_skip(self):
            src, src_file = _setup_dir_and_file()
            dest, dest_file = _setup_dir_and_file()
            make_dir(f'{dest}/subdir')

            move_dir(src, dest, dir_exists='skip', file_exists='replace')
            assert Path(dest_file).read_text() != Path(src_file).read_text()
            assert Path(f'{dest}/subdir').is_dir()

        def test_file_exists_error(self):
            with pytest.raises(FileExistsError):
                src = _setup_tree()
                dest = _setup_tree()
                move_dir(src, dest, dir_exists='merge', file_exists='error')

        def test_file_exists_replace(self):
            src, src_file = _setup_dir_and_file()
            dest, dest_file = _setup_dir_and_file()
            src_file_text = Path(src_file).read_text()

            move_dir(src, dest, dir_exists='merge', file_exists='replace')
            assert Path(dest_file).read_text() == src_file_text

        def test_file_exists_skip(self):
            src, src_file = _setup_dir_and_file()
            dest, dest_file = _setup_dir_and_file()

            src_file_text = Path(src_file).read_text()
            original = Path(dest_file).read_text()

            move_dir(src, dest, dir_exists='merge', file_exists='skip')
            assert Path(dest_file).read_text() != src_file_text
            assert Path(dest_file).read_text() == original

        def test_make_dirs_false(self):
            with pytest.raises(FileNotFoundError):
                with tempfile.TemporaryDirectory() as d:
                    src = get_unique_path(d + '/{}_src')
                    dest = get_unique_path(d + '/{}_dest')
                    make_dir(src)
                    move_dir(src, dest, make_dirs=False)

        def test_make_dirs_true(self):
            with tempfile.TemporaryDirectory() as d:
                src = _setup_tree()
                src_name = Path(src).name
                src_copy = Path(get_unique_path('{}_src_copy'))
                copy_dir(src, src_copy)

                dest = get_unique_path(d + '/{}_dest')
                move_dir(src, dest, make_dirs=True)
                _check_tree(src_copy/src_name, f'{dest}/{src_name}')

                src = _setup_tree()
                src_name = Path(src).name
                src_copy = Path(get_unique_path('{}_src_copy'))
                copy_dir(src, src_copy)

                dest_2 = get_unique_path(d + '/{}_dest_2')
                move_dir(src, f'{dest_2}/{src_name}', make_dirs=True)
                _check_tree(src_copy/src_name, f'{dest_2}/{src_name}')

        def test_same_path(self):
            with pytest.raises(ValueError):
                src = _setup_dir()
                move_dir(src, src)

        def test_src_is_file(self):
            with pytest.raises(NotADirectoryError):
                src = _setup_file()
                dest = get_unique_path('{}_new')
                move_dir(src, dest)

        def test_src_is_missing(self):
            with pytest.raises(FileNotFoundError):
                src = get_unique_path()
                move_dir(src, f'{src}_copy')

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_dest(self, mistake):
            with pytest.raises(TypeError):
                move_dir('src', mistake)

        @pytest.mark.parametrize('mistake', [0, None, {}, True, 'hello'])
        def test_precondition_dir_exists(self, mistake):
            with pytest.raises(ValueError):
                move_dir('src', 'dest', dir_exists=mistake)

        @pytest.mark.parametrize('mistake', [0, None, {}, 'hello'])
        def test_precondition_file_exists(self, mistake):
            with pytest.raises(ValueError):
                move_dir('src', 'dest', file_exists=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_make_dirs(self, mistake):
            with pytest.raises(TypeError):
                src, dest = _setup_dir(n=2)
                move_dir(src, dest, make_dirs=mistake)

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_src(self, mistake):
            with pytest.raises(TypeError):
                # noinspection PyTypeChecker
                move_dir(mistake, 'dest')

    # noinspection PyTypeChecker
    class Test_move_file:
        def test_dir(self):
            with pytest.raises(IsADirectoryError):
                src = _setup_dir()
                dest = get_unique_path()
                move_file(src, dest)

        def test_exists_error(self):
            with pytest.raises(FileExistsError):
                src, dest = _setup_file(n=2)
                move_file(src, dest, exists='error')

        def test_exists_replace(self):
            src, dest = _setup_file(n=2)
            expected_text = Path(src).read_text()
            move_file(src, dest, exists='replace')

            text = Path(dest).read_text()
            assert text == expected_text
            assert text != dest
            assert not Path(src).exists()

        def test_exists_skip(self):
            src, dest = _setup_file(n=2)
            move_file(src, dest, exists='skip')

            text = Path(dest).read_text()
            assert text != Path(src).read_text()
            assert text == dest
            assert Path(src).exists()

        def test_make_dirs_false(self):
            with pytest.raises(FileNotFoundError):
                src = _setup_file()
                dest = get_unique_path('{}/a/b/c/')
                move_file(src, dest, make_dirs=False)

        def test_make_dirs_true(self):
            src = _setup_file()
            dest = _setup_path()
            move_file(src, dest, make_dirs=True)
            assert Path(dest).read_text() == src
            assert not Path(src).exists()

        def test_same_path(self):
            with pytest.raises(ValueError):
                src = _setup_file()
                move_file(src, src)

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_src(self, mistake):
            with pytest.raises(TypeError):
                move_file(mistake, 'a')

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_precondition_dest(self, mistake):
            with pytest.raises(TypeError):
                move_file('a', mistake)

        @pytest.mark.parametrize('mistake', all_types_besides('booleans'))
        def test_precondition_make_dirs(self, mistake):
            with pytest.raises(TypeError):
                src, dest = _setup_file(n=2)
                move_file(src, dest, make_dirs=mistake)

        @pytest.mark.parametrize('mistake', [0, None, {}, 'hello'])
        def test_precondition_exists(self, mistake):
            src, dest = _setup_file(n=2)
            with pytest.raises(ValueError):
                move_file(src, dest, exists=mistake)


class TestGroup_paths:
    class Test_get_unique_path:
        def test_consecutive_values(self):
            root = Path(_setup_dir())
            for i in range(1, 11):
                file = get_unique_path(root/'{}')
                assert file == Path(f'{root}/{i}')
                Path(file).touch()

        @pytest.mark.parametrize('mistake', all_types_besides(['str', 'path']))
        def test_invalid_pattern_type(self, mistake):
            with pytest.raises(TypeError):
                get_unique_path(mistake)

        @pytest.mark.parametrize('mistake', ['', '{0}', '{a}', '{:f}'])
        def test_invalid_placeholder(self, mistake):
            with pytest.raises(ValueError):
                get_unique_path(mistake)

        def test_keeps_slash_at_the_end(self):
            assert get_unique_path('{}/').endswith('/')

        def test_returns_same_type_as_input(self):
            assert isinstance(get_unique_path('{}'), str)
            assert isinstance(get_unique_path(Path('{}')), Path)

        @pytest.mark.parametrize('mistake', ['hello_{}', '{}', './{}', '/a/{}', '/{}/a', '{:06d}'])
        def test_not_exists(self, mistake):
            path = get_unique_path(mistake)
            assert not Path(path).exists()


def _check_tree(src, dest):
    for src_item in Path(src).iterdir():
        dest_item = str(src_item).replace(str(src), str(dest))
        dest_item = Path(dest_item)

        if src_item.is_dir():
            assert dest_item.is_dir()
            _check_tree(src_item, dest_item)
        else:
            assert dest_item.is_file()


def _setup_tree(path=None):
    path = _setup_path(path)

    for d in [path, f'{path}/a', f'{path}/b', f'{path}/b/c']:
        _setup_dir(d)

    for f in [f'{path}/1.txt', f'{path}/a/2.txt', f'{path}/b/3.txt', f'{path}/b/c/4.txt']:
        _setup_file(f)

    return str(path)


def _setup_tree_2(path=None):
    path = _setup_tree(path)

    make_dir(f'{path}/a/a')
    make_dir(f'{path}/b/a')
    make_dir(f'{path}/b/c/a')
    make_dir(f'{path}/b/c/a/d')

    Path(f'{path}/1.py').touch()
    Path(f'{path}/a/2.py').touch()
    Path(f'{path}/b/a/5.py').touch()
    Path(f'{path}/a/a/6.py').touch()

    return path


def _setup_dir_and_file(path_dir=None, filename=None):
    if filename is None:
        filename = 'file'

    dir_ = _setup_dir(path_dir)
    file = _setup_file(Path(dir_)/filename)
    return dir_, file


def _setup_file(path=None, n=1):
    files = []
    for _ in range(n):
        file = _setup_path(path)
        make_dir(Path(file).parent)
        Path(file).write_text(file)

        files.append(file)

    if n == 1:
        return files[0]
    else:
        return files


def _setup_dir(path=None, n=1):
    dirs = []
    for _ in range(n):
        path = _setup_path(path)
        make_dir(path)
        dirs.append(path)

    if n == 1:
        return dirs[0]
    else:
        return dirs


def _setup_path(path=None):
    if path is None:
        return get_unique_path()
    elif '{}' in str(path):
        return get_unique_path(path)
    else:
        return str(path)



import itertools

import pytest as pytest

from katalytic.data import all_types_besides
# noinspection PyProtectedMember
from katalytic.maths.bboxes import (
    _FORMATS, calc_bbox_area, calc_bbox_center, calc_IoB, calc_IoB_max, calc_IoB_min,
    calc_IoU, convert_bbox, non_max_suppression, set_bbox_scores, intersect_bboxes, is_bbox)


_EXAMPLES = {
    'xyXY': (1,2,9,9),
    'xyXY_s': ((1,2,9,9), 0.5),
    'xyXYs': (1,2,9,9,0.5),
    'xy_XY': ((1,2), (9,9)),
    'xy_XY_s': ((1,2), (9,9), 0.5),
    'xy_wh': ((1,2), (8,7)),
    'xy_wh_s': ((1,2), (8,7), 0.5),
    'xywh': (1,2,8,7),
    'xywh_s': ((1,2,8,7), 0.5),
    'xywhs': (1,2,8,7,0.5),
}


def _check_all_besides(bbox, blacklist):
    formats = set(_FORMATS) - set(blacklist)
    return list(itertools.product([bbox], formats))


def _pick_all_besides(blacklist):
    return [bbox for fmt, bbox in _EXAMPLES.items() if fmt not in blacklist]


def _generate_all_possible_conversions():
    all_combinations = itertools.product(_EXAMPLES.items(), _EXAMPLES.items())
    possible = []
    for (fmt_1, box_1), (fmt_2, box_2) in all_combinations:
        if fmt_2.endswith('s') and not fmt_1.endswith('s'):
            continue
        elif fmt_2 == fmt_1:
            continue

        possible.append((box_1, fmt_1, fmt_2, box_2))

    return possible


def _generate_all_impossible_conversions():
    all_combinations = itertools.product(_EXAMPLES.items(), _EXAMPLES.items())
    impossible = []
    for (fmt_1, box_1), (fmt_2, box_2) in all_combinations:
        if fmt_2.endswith('s') and not fmt_1.endswith('s'):
            impossible.append((box_1, fmt_1, fmt_2))

    return impossible


class Test_calc_area:
    @pytest.mark.parametrize('fmt, bbox', _EXAMPLES.items())
    def test_all_formats(self, fmt, bbox):
        assert calc_bbox_area(bbox, fmt) == 8 * 7

    @pytest.mark.parametrize('bbox, fmt, area', [
        ((0, 0, 5, 5), 'xyXY', 25),
        ((0, 0, 10, 10), 'xywh', 100),
        ((0, 0, 1, 5), 'xyXY', 5),
        ((0, 0, 10, 1), 'xywh', 10),
    ])
    def test_basic(self, bbox, fmt, area):
        assert calc_bbox_area(bbox, fmt) == area


class Test_convert_bbox:
    @pytest.mark.parametrize('wrong_type', all_types_besides('sequences'))
    def test_wrong_type(self, wrong_type):
        with pytest.raises(TypeError):
            convert_bbox(wrong_type, 'xyXY', 'xywh')

    @pytest.mark.parametrize('bbox, before, after, _', _generate_all_possible_conversions())
    def test_not_the_specified_format(self, bbox, before, after, _):
        # skip false positives
        if 'wh' in before and 'XY' in after:
            return
        elif 'XY' in before and 'wh' in after:
            return

        with pytest.raises(ValueError):
            convert_bbox(bbox, after, before)

    @pytest.mark.parametrize('fmt, bbox', _EXAMPLES.items())
    def test_same_format(self, fmt, bbox):
        with pytest.raises(ValueError):
            convert_bbox(bbox, fmt, fmt)

    @pytest.mark.parametrize('wrong_format', [1, None, 'asd', 'qwerty'])
    def test_wrong_format(self, wrong_format):
        with pytest.raises(ValueError):
            convert_bbox((1,2,3,4), 'xywh', wrong_format)

        with pytest.raises(ValueError):
            convert_bbox((1,2,3,4), wrong_format, 'xywh')

    @pytest.mark.parametrize('bbox, before, after', _generate_all_impossible_conversions())
    def test_impossible(self, bbox, before, after):
        with pytest.raises(ValueError):
            convert_bbox(bbox, before, after)

    @pytest.mark.parametrize('bbox, before, after, expected', _generate_all_possible_conversions())
    def test_expected(self, bbox, before, after, expected):
        assert convert_bbox(bbox, before, after) == expected


class Test_calc_bbox_center:
    @pytest.mark.parametrize('bbox, fmt, expected', [
        ((0, 0, 5, 4), 'xyXY', (2.5, 2)),
        ((0, 0, 3, 10), 'xywh', (1.5, 5.0)),
    ])
    def test_float(self, bbox, fmt, expected):
        assert calc_bbox_center(bbox, fmt) == expected

    @pytest.mark.parametrize('bbox, fmt, expected', [
        ((0, 0, 5, 4), 'xyXY', (2, 2)),
        ((0, 0, 3, 10), 'xywh', (1, 5)),
    ])
    def test_int(self, bbox, fmt, expected):
        assert calc_bbox_center(bbox, fmt, as_int=True) == expected


class Test_calc_IoU:
    @pytest.mark.parametrize('bbox, fmt_1, bbox_2, fmt_2', [
        ((0, 0, 5, 5), 'xyXY', (5, 5, 10, 10), 'xyXY'),
        ((1, 1, 4, 4), 'xywh', (6, 6, 9, 9), 'xyXY'),
    ])
    def test_no_overlap(self, bbox, fmt_1, bbox_2, fmt_2):
        assert calc_IoU(bbox, fmt_1, bbox_2, fmt_2) == 0

    @pytest.mark.parametrize('bbox_1, fmt_1, bbox_2, fmt_2', [
        ((0, 0, 5, 5), 'xyXY', (0, 0, 5, 5), 'xyXY'),
        ((1, 1, 4, 4), 'xywh', (1, 1, 5, 5), 'xyXY'),
        (((1, 1), (4, 4), 0.3), 'xy_wh_s', ((1, 1), (5, 5), 0.6), 'xy_XY_s'),
    ])
    def test_perfect_overlap(self, bbox_1, fmt_1, bbox_2, fmt_2):
        assert calc_IoU(bbox_1, fmt_1, bbox_2, fmt_2) == 1

    @pytest.mark.parametrize('bbox_1, fmt_1, bbox_2, fmt_2, expected', [
        ((0, 0, 5, 5), 'xyXY', (3, 4, 7, 7), 'xyXY', 0.05714285714285714),
        ((0, 0, 5, 5), 'xyXY', (1, 1, 6, 6), 'xyXY', 0.47058823529411764),
        (((0, 0, 5, 5), 0.9), 'xyXY_s', (0, 1, 10, 3), 'xywh', 0.375),
    ])
    def test_partial_overlap(self, bbox_1, fmt_1, bbox_2, fmt_2, expected):
        assert calc_IoU(bbox_1, fmt_1, bbox_2, fmt_2) == expected


class Test_calc_IoB:
    @pytest.mark.parametrize('bbox, fmt_1, bbox_2, fmt_2', [
        ((0, 0, 5, 5), 'xyXY', (5, 5, 10, 10), 'xyXY'),
        ((1, 1, 4, 4), 'xywh', (6, 6, 9, 9), 'xyXY'),
    ])
    def test_no_overlap(self, bbox, fmt_1, bbox_2, fmt_2):
        assert calc_IoB(bbox, fmt_1, bbox_2, fmt_2) == 0

    @pytest.mark.parametrize('bbox_1, fmt_1, bbox_2, fmt_2', [
        ((0, 0, 5, 5), 'xyXY', (0, 0, 5, 5), 'xyXY'),
        ((1, 1, 4, 4), 'xywh', (1, 1, 5, 5), 'xyXY'),
        (((1, 1), (4, 4), 0.3), 'xy_wh_s', ((1, 1), (5, 5), 0.6), 'xy_XY_s'),
    ])
    def test_perfect_overlap(self, bbox_1, fmt_1, bbox_2, fmt_2):
        assert calc_IoB(bbox_1, fmt_1, bbox_2, fmt_2) == 1

    @pytest.mark.parametrize('bbox_1, fmt_1, bbox_2, fmt_2, expected', [
        ((0, 0, 5, 5), 'xyXY', (3, 4, 7, 7), 'xyXY', 0.08),
        ((0, 0, 5, 5), 'xyXY', (1, 1, 6, 6), 'xyXY', 0.64),
        (((0, 0, 5, 5), 0.9), 'xyXY_s', (0, 1, 10, 3), 'xywh', 0.6),
    ])
    def test_partial_overlap(self, bbox_1, fmt_1, bbox_2, fmt_2, expected):
        assert calc_IoB(bbox_1, fmt_1, bbox_2, fmt_2) == expected


class Test_calc_IoB_min:
    @pytest.mark.parametrize('bbox_1, fmt_1, bbox_2, fmt_2, expected', [
        ((0, 0, 5, 5), 'xyXY', (3, 4, 7, 7), 'xyXY', 0.08),
        ((0, 0, 5, 5), 'xyXY', (1, 1, 6, 6), 'xyXY', 0.64),
        (((0, 0, 5, 5), 0.9), 'xyXY_s', (0, 1, 10, 3), 'xywh', 0.5),
    ])
    def test_partial_overlap(self, bbox_1, fmt_1, bbox_2, fmt_2, expected):
        assert calc_IoB_min(bbox_1, fmt_1, bbox_2, fmt_2) == expected


class Test_calc_IoB_max:
    @pytest.mark.parametrize('bbox_1, fmt_1, bbox_2, fmt_2, expected', [
        ((0, 0, 5, 5), 'xyXY', (3, 4, 7, 7), 'xyXY', 0.16666666666666666),
        ((0, 0, 5, 5), 'xyXY', (1, 1, 6, 6), 'xyXY', 0.64),
        (((0, 0, 5, 5), 0.9), 'xyXY_s', (0, 1, 10, 3), 'xywh', 0.6),
    ])
    def test_partial_overlap(self, bbox_1, fmt_1, bbox_2, fmt_2, expected):
        assert calc_IoB_max(bbox_1, fmt_1, bbox_2, fmt_2) == expected


class Test_set_bbox_scores:
    @pytest.mark.parametrize('after', ['xywh', 'xy_XY', 'xy_wh'])
    def test_invalid_after_format(self, after):
        with pytest.raises(ValueError):
            set_bbox_scores([(1,2,3,4)], [0], 'xyXY', after)

    @pytest.mark.parametrize('scores', [-1, -0.1, 1.1, 2, 3.5])
    def test_invalid_scores(self, scores):
        with pytest.raises(ValueError):
            set_bbox_scores([(1,2,3,4)], [scores], 'xyXY', 'xyXYs')

    @pytest.mark.parametrize('bboxes, scores', [
        ([(1,2,3,4)], [1, 0.2]),
        ([(1,2,3,4), (1,2,3,4)], [1]),
    ])
    def test_unequal_lengths(self, bboxes, scores):
        with pytest.raises(ValueError):
            set_bbox_scores(bboxes, scores, 'xyXY', 'xyXYs')

    @pytest.mark.parametrize('bboxes, scores, before, after, expected', [
        (
            [((1,2),(3,4))],
            [1],
            'xy_XY',
            'xyXYs',
            [(1,2,3,4,1)]
        ),
        (
            [((1,2),(3,4)), ((2,2),(3,4))],
            [1, 0.2],
            'xy_XY',
            'xyXYs',
            [(1,2,3,4,1), (2,2,3,4,0.2)]
        ),
        (
            [((1,2,3,4)), ((2,2,3,4))],
            [1, 0.2],
            'xyXY',
            'xyXY_s',
            [((1,2,3,4), 1), ((2,2,3,4), 0.2)]
        ),
    ])
    def test_ok(self, bboxes, scores, before, after, expected):
        assert set_bbox_scores(bboxes, scores, before, after) == expected


class Test_non_max_suppression:
    def test_no_scores(self):
        with pytest.raises(ValueError):
            non_max_suppression([(1,2,3,4)], 'xyXY')

    def test_scores(self):
        bboxes = [(1,2,3,4), (5,6,7,8), (5,6,7,7)]
        expected = [(1,2,3,4), (5,6,7,8)]
        scores = [0.9, 0.8, 0.7]
        assert non_max_suppression(bboxes, 'xyXY', scores=scores) == expected

    def test_empty(self):
        assert non_max_suppression([], 'xyXYs') == []

    def test_perfect_overlap(self):
        bboxes = [
            ((10, 10, 100, 100), 0.9),
            ((10, 10, 100, 100), 0.8),
            ((10, 10, 100, 100), 0.7)
        ]

        assert non_max_suppression(bboxes, 'xyXY_s') == [((10, 10, 100, 100), 0.9)]

    def test_no_overlap(self):
        bboxes = expected = [
            [10, 10, 50, 50, 0.9],
            [60, 60, 100, 100, 0.8]
        ]

        assert non_max_suppression(bboxes, 'xyXYs') == expected

    def test_partial_overlap(self):
        bboxes = [
            [10, 10, 100, 100, 0.9],
            [100, 100, 190, 190, 0.6],
            [100, 100, 185, 185, 0.7],
            [90, 90, 190, 190, 0.7],
            [50, 50, 150, 150, 0.8],
        ]

        assert non_max_suppression(bboxes, 'xyXYs') == [
            [10, 10, 100, 100, 0.9],
            [50, 50, 150, 150, 0.8],
            [100, 100, 185, 185, 0.7],
        ]


class Test_intersect_bboxes:
    @pytest.mark.parametrize('bbox_1, fmt_1, bbox_2, fmt_2', [
        ((0, 0, 5, 5), 'xyXY', (5, 5, 10, 10), 'xyXY'),
        ((1, 1, 4, 4), 'xywh', (6, 6, 9, 9), 'xyXY'),
    ])
    def test_no_overlap(self, bbox_1, fmt_1, bbox_2, fmt_2):
        assert intersect_bboxes(bbox_1, fmt_1, bbox_2, fmt_2) is None

    @pytest.mark.parametrize('bbox_1, fmt_1, bbox_2, fmt_2', [
        ((0, 0, 5, 5), 'xyXY', (0, 0, 5, 5), 'xyXY'),
        ((1, 1, 4, 4), 'xywh', (1, 1, 5, 5), 'xyXY'),
        (((1, 1), (4, 4), 0.3), 'xy_wh_s', ((1, 1), (5, 5), 0.6), 'xy_XY_s'),
    ])
    def test_perfect_overlap(self, bbox_1, fmt_1, bbox_2, fmt_2):
        assert intersect_bboxes(bbox_1, fmt_1, bbox_2, fmt_2) == bbox_1

    @pytest.mark.parametrize('bbox_1, fmt_1, bbox_2, fmt_2, expected', [
        ((0, 0, 5, 5), 'xyXY', (3, 4, 7, 7), 'xyXY', (3, 4, 5, 5)),
        (((0, 0, 5, 5), 0.9), 'xyXY_s', (0, 1, 10, 3), 'xywh', ((0, 1, 5, 4), 0.9)),
    ])
    def test_partial_overlap(self, bbox_1, fmt_1, bbox_2, fmt_2, expected):
        assert intersect_bboxes(bbox_1, fmt_1, bbox_2, fmt_2) == expected


class Test_is_bbox:
    @pytest.mark.parametrize('wrong_type', all_types_besides('sequences'))
    def test_wrong_format(self, wrong_type):
        with pytest.raises(TypeError):
            is_bbox(wrong_type, 'xyXY')

    @pytest.mark.parametrize('fmt', [1, None, 'asd', 'qwerty'])
    def test_wrong_format(self, fmt):
        with pytest.raises(ValueError):
            is_bbox((1,2,3,4), fmt)

    @pytest.mark.parametrize('fmt, bbox', _EXAMPLES.items())
    def test_True(self, fmt, bbox):
        assert is_bbox(bbox, fmt)

    @pytest.mark.parametrize('bbox, fmt', [
        *_check_all_besides((), []),
        *_check_all_besides(range(1), []),
        *_check_all_besides(range(2), []),
        *_check_all_besides(range(3), []),
        *_check_all_besides((*range(4), -0.1), []),
        *_check_all_besides((*range(4), 1.1), []),
        *_check_all_besides((*range(4), 2), []),
        *_check_all_besides(range(6), []),
        *_check_all_besides(((1,2),(3,4),(5,6),(7,8)), []),

        *_check_all_besides(((1,2), (1,1)), ['xy_wh']),
        *_check_all_besides(((1,2), (1,1), 0.5), ['xy_wh_s']),
        *_check_all_besides((1,2,1,1), ['xywh']),
        *_check_all_besides(((1,2,1,1), 0.5), ['xywh_s']),
        *_check_all_besides((1,2,1,1,0.5), ['xywhs']),

        # There's no way to check correctly against the "wh" equivalents
        # They would return false positives and make the test fail
        *_check_all_besides((1,2,3,4), ['xyXY', 'xywh']),
        *_check_all_besides(((1,2,3,4), 0.5), ['xyXY_s', 'xywh_s']),
        *_check_all_besides((1,2,3,4,0.5), ['xyXYs', 'xywhs']),
        *_check_all_besides(((1,2), (3,4)), ['xy_XY', 'xy_wh']),
        *_check_all_besides(((1,2), (3,4), 0.5), ['xy_XY_s', 'xy_wh_s']),
    ])
    def test_False(self, bbox, fmt):
        assert not is_bbox(bbox, fmt)
